import os
import json
import h5py
import itertools
import numpy as np
from pathlib import Path


class ExperimentSettings:
    """Mother class to control experiment parameters

    This class is responsible for the following

    - Defining paths and model names
    - Choosing the device on which to run computations
    - Specifying all hyperparameters such as model configuration, datasets, features etc

    Args:
        cli_args (argparse.Namespace) command line arguments
    """

    def __init__(self, cli_args, action=None):

        self.cli_args = {}

        # Ensure cli_args is of type dict
        if not isinstance(cli_args, dict):
            cli_args = vars(cli_args)

        # Update and overwrite options
        self.__dict__.update(cli_args)
        self.cli_args.update(cli_args)

        self.action = action

        self.device = "cpu"
        if self.use_cuda:
            self.device = "cuda"

        if self.model == "variational":
            self.weight_decay = self.weight_decay
        else:
            self.weight_decay = 0.0

        # Load simulation and training settings and prepare directories
        if self.no_dump:
            pass
        else:
            self._setup_dir()
            # Set the database file names
            self._set_database_file_names()
            # Set the feature lists
            if "all_features" not in cli_args:
                self._set_feature_lists()

            self.overwrite = not self.no_overwrite

            # filter combination
            list_filters_combination = []
            for i in range(1, len(self.list_filters) + 1):
                tmp = [
                    "".join(t)
                    for t in list(itertools.combinations(self.list_filters, i))
                ]
                list_filters_combination = list_filters_combination + tmp
            self.list_filters_combination = list_filters_combination
            # action: make_data
            if self.action == "make_data":
                self._save_to_json(self.processed_dir)
            # other actions: train_rnn, validate_rnn, show, performance or not specified
            if self.action != "make_data":
                self.set_training_setting()

    def _setup_dir(self):
        """Configure directories where data is read from or dumped to
        during the course of an experiment
        """

        for path in [
            # f"{self.raw_dir}",
            # f"{self.fits_dir}",
            f"{self.dump_dir}/explore",
            # f"{self.dump_dir}/stats",
            f"{self.dump_dir}/figures",
            f"{self.dump_dir}/lightcurves",
            # f"{self.dump_dir}/latex",
            f"{self.dump_dir}/processed",
            f"{self.dump_dir}/preprocessed",
            f"{self.dump_dir}/models",
        ]:

            setattr(self, Path(path).name + "_dir", path)

            Path(path).mkdir(exist_ok=True, parents=True)

    def _set_database_file_names(self):
        """Create a unique database name based on the dataset required
        by the settings
        """

        out_file = f"{self.processed_dir}/database"
        self.pickle_file_name = out_file + ".pickle"
        self.hdf5_file_name = out_file + ".h5"

    def _set_feature_lists(self):
        """Utility to define the features used to train NN=based models"""

        self.training_features_to_normalize = [
            f"FLUXCAL_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += [
            f"FLUXCALERR_{f}" for f in self.list_filters
        ]
        self.training_features_to_normalize += ["delta_time"]

    def _add_feature_lists(self):
        # If the database has been created, add the list of all features
        with h5py.File(self.hdf5_file_name, "r") as hf:
            self.all_features = hf["features"][:].astype(str)

            self.non_redshift_features = [
                f
                for f in self.all_features
                if "FLUX" in f
                or f in self.list_filters_combination
                or "delta_time" in f
            ]

            # Optionally add redshift
            self.redshift_features = []
            if self.redshift == "zpho":
                self.redshift_features = [
                    f for f in self.all_features if "HOSTGAL_PHOTOZ" in f
                ]
            elif self.redshift == "zspe":
                self.redshift_features = [
                    f for f in self.all_features if "HOSTGAL_SPECZ" in f
                ]

            self.training_features = self.non_redshift_features + self.redshift_features

            if self.additional_train_var:
                self.training_features += [
                    k
                    for k in self.additional_train_var
                    if k not in self.training_features
                ]

    def _save_to_json(self, save_to_dir):
        d_tmp = {}
        for k, v in self.__dict__.items():
            if isinstance(v, np.ndarray):
                v = v.tolist()
            d_tmp[k] = v

        os.makedirs(save_to_dir, exist_ok=True)
        # Dump the command line arguments (for model restoration)
        with open(Path(save_to_dir) / "cli_args.json", "w") as f:
            json.dump(d_tmp, f, indent=4, sort_keys=True)

    def _set_pytorch_model_name(self):
        """Define the model name for all NN based classifiers"""
        name = f"{self.model}_S_{self.seed}_CLF_{self.nb_classes}"
        name += f"_R_{self.redshift}"
        name += f"_{self.source_data}_DF_{self.data_fraction}_N_{self.norm}"
        name += f"_{self.layer_type}_{self.hidden_dim}x{self.num_layers}"
        name += f"_{self.dropout}"
        name += f"_{self.batch_size}"
        name += f"_{self.bidirectional}"
        name += f"_{self.rnn_output_option}"
        if "bayesian" in self.model:
            name += (
                f"_Bayes_{self.pi}_{self.log_sigma1}_{self.log_sigma2}"
                f"_{self.rho_scale_lower}_{self.rho_scale_upper}"
                f"_{self.log_sigma1_output}_{self.log_sigma2_output}"
                f"_{self.rho_scale_lower_output}_{self.rho_scale_upper_output}"
            )
        if self.cyclic:
            name += "_C"
        if self.weight_decay > 0:
            name += f"_WD_{self.weight_decay}"

        self.pytorch_model_name = name
        self.rnn_dir = f"{self.models_dir}/{self.pytorch_model_name}"

        # deserializing numpy arrays to save as json
        d_tmp = {}
        for k, v in self.__dict__.items():
            if isinstance(v, np.ndarray):
                v = v.tolist()
            d_tmp[k] = v

        if self.action == "train_rnn":
            self._save_to_json(self.rnn_dir)

    def _load_normalization(self):
        """Create an array holding the data-normalization parameters
        used to normalize certain features in the NN-based classification
        pipeline
        """
        self.idx_features = [
            i for (i, f) in enumerate(self.all_features) if f in self.training_features
        ]

        self.idx_specz = [
            i for (i, f) in enumerate(self.training_features) if "HOSTGAL_SPECZ" in f
        ]

        self.idx_flux = [
            i for (i, f) in enumerate(self.training_features) if "FLUXCAL_" in f
        ]

        self.idx_fluxerr = [
            i for (i, f) in enumerate(self.training_features) if "FLUXCALERR_" in f
        ]

        self.idx_delta_time = [
            i for (i, f) in enumerate(self.training_features) if "delta_time" in f
        ]

        self.idx_features_to_normalize = [
            i
            for (i, f) in enumerate(self.all_features)
            if f in self.training_features_to_normalize
            or f.replace("DES-", "")
            in self.training_features_to_normalize  # Horrible fix for new SNANA format
        ]

        self.d_feat_to_idx = {f: i for i, f in enumerate(self.all_features)}

        list_norm = []

        with h5py.File(self.hdf5_file_name, "r") as hf:

            for f in self.training_features_to_normalize:

                if self.norm == "perfilter":

                    minv = np.array(hf[f"normalizations/{f}/min"])
                    meanv = np.array(hf[f"normalizations/{f}/mean"])
                    stdv = np.array(hf[f"normalizations/{f}/std"])

                    list_norm.append([minv, meanv, stdv])

                else:

                    if "FLUX" in f:
                        prefix = f.split("_")[0]
                        minv = np.array(hf[f"normalizations_global/{prefix}/min"])
                        meanv = np.array(hf[f"normalizations_global/{prefix}/mean"])
                        stdv = np.array(hf[f"normalizations_global/{prefix}/std"])
                    else:
                        minv = np.array(hf[f"normalizations/{f}/min"])
                        meanv = np.array(hf[f"normalizations/{f}/mean"])
                        stdv = np.array(hf[f"normalizations/{f}/std"])

                    list_norm.append([minv, meanv, stdv])

        self.arr_norm = np.array(list_norm)

    def set_training_setting(self):
        """Init settings needed for training and the following actions"""
        if "all_features" not in self.cli_args.keys():
            self._add_feature_lists()

        self._set_pytorch_model_name()
        # Get the feature normalization dict
        self._load_normalization()

    def check_data_exists(self):
        """Utility to check the database has been built"""

        database_file = f"{self.processed_dir}/database.h5"
        assert os.path.isfile(database_file)



import os
import pandas as pd
from pathlib import Path
import supernnova.conf as conf
from . import superNNova_plots as sp
from . import superNNova_metrics as sm
from ..utils import logging_utils as lu
from ..visualization import early_prediction

"""
Obtaining metrics and plots for SuperNNova paper

Selection of models is hard coded

Code is far from optimized
"""

"""
Best performing algorithms in SuperNNova
"""
Base = (
    "DES_vanilla_CLF_2_R_None_photometry_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C"
)
Var = "DES_variational_CLF_2_R_None_photometry_DF_1.0_N_global_lstm_32x2_0.01_128_True_mean_C_WD_1e-07"
BBB = "DES_bayesian_CLF_2_R_None_photometry_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_Bayes_0.75_-1.0_-7.0_4.0_3.0_-0.5_-0.1_3.0_2.0"  # _KL_0.1"
RF = "DES_randomforest_CLF_2_R_None_saltfit_DF_1.0_N_global"
list_models = [RF, Base, Var, BBB]
list_models_rnn = [Base, Var, BBB]

# useful formats
Base_salt = Base.replace("photometry", "saltfit")


def SuperNNova_stats_and_plots(settings):
    """Reproduce stats and plots used for SuperNNova paper.
    BEWARE: Selection is hardcoded

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    # Load summary statistics
    df_stats = pd.read_csv(Path(settings.stats_dir) / "summary_stats.csv")

    # Create latex tables
    # sm.create_accuracy_latex_tables(df_stats, settings)

    # Rest of stats and plots in paper
    # can be ran in debug mode: only printing model names
    # or in no plot mode: only printing stats
    SuperNNova_stats_and_plots_thread(df_stats, settings, plots=True, debug=False)


def SuperNNova_stats_and_plots_thread(df, settings, plots=True, debug=False):
    """Stats quoted in paper which are not in the latex tables and plots

    Args:
        df (pandas.DataFrame) : summary statistics df
        settings (ExperimentSettings): custom class to hold hyperparameters
        plots (Boolean optional): make pltos or only printout stats
        debug (Boolean optional): only print tasks
    Returns:
        printout: stats as organized in paper
        figures (png) : figures for paper at settings.dump_dir/figures/
        lightcurves (png): lightcurves used on paper at settings.dump_dir/lightcurves/modelname.*png
    """

    """
    Ordered as in paper
    """
    pd.set_option("max_colwidth", 1000)
    print(lu.str_to_greenstr("STATISTICS USED IN SUPERNNOVA"))

    # Baseline experiments
    baseline(df, settings, plots, debug)
    # Bayesian experiments
    df_delta, df_delta_ood = sm.get_delta_metrics(df, settings)
    bayesian(df, df_delta, df_delta_ood, settings, plots, debug)
    # Towards statistical analyses/cosmology
    towards_cosmo(df, df_delta, df_delta_ood, settings, plots, debug)


def baseline(df, settings, plots, debug):
    """
    Baseline RNN
    """
    # 0. Figure example
    if plots:
        print(
            lu.str_to_yellowstr("Plotting candidates for Baseline binary (Figure 2.)")
        )
        model_file = f"{settings.models_dir}/{Base.replace('DES_vanilla_','vanilla_S_0_')}/{Base.replace('DES_vanilla_','vanilla_S_0_')}.pt"
        if os.path.exists(model_file):
            if debug:
                print(model_file)
            else:
                model_settings = conf.get_settings_from_dump(model_file)
                early_prediction.make_early_prediction(
                    model_settings, nb_lcs=20, do_gifs=True
                )
        else:
            print(lu.str_to_redstr(f"File not found {model_file}"))

    # 1. Hyper-parameters
    # saltfit, DF 0.2
    sel_criteria = ["DES_vanilla_CLF_2_R_None_saltfit_DF_0.2"]
    print(lu.str_to_bluestr(f"Hyperparameters {sel_criteria}"))
    if not debug:
        sm.get_metric_ranges(df, sel_criteria)

    # 2. Normalization
    # saltfit, DF 0.5
    sel_criteria = Base_salt.replace("DF_1.0", "DF_0.5").split("global")
    print(lu.str_to_bluestr(f"Normalization {sel_criteria}"))
    if not debug:
        sm.get_metric_ranges(df, sel_criteria)

    # 3. Comparing with other methods
    print(lu.str_to_bluestr("Other methods:"))
    # Figure: accuracy vs. number of SNe
    if plots:
        print(lu.str_to_yellowstr("Plotting accuracy vs. SNe (Figure 3.)"))
        if not debug:
            sp.performance_plots(settings)
    # baseline best, saltfit
    if debug:
        print(Base_salt)
        print(Base_salt.replace("DF_1.0", "DF_0.05"))
        print(Base_salt.replace("DF_1.0", "DF_0.05").replace("None", "zpho"))
    else:
        sm.acc_auc_df(df, [Base_salt], data="saltfit")

        # RF and baseline comparisson with Charnock Moss
        sm.acc_auc_df(
            df,
            [
                RF,
                Base_salt.replace("DF_1.0", "DF_0.05"),
                Base_salt.replace("DF_1.0", "DF_0.05").replace("None", "zpho"),
            ],
        )

    # 4. Redshift, contamination
    # baseline saltfit, 1.0, all redshifts
    sel_criteria = Base_salt.split("None")
    print("salt")
    if debug:
        print(sel_criteria, Base.split("None"))
    if not debug:
        sm.print_contamination(df, sel_criteria, settings, data="saltfit")
    print("photometry")
    if not debug:
        sm.print_contamination(df, Base.split("None"), settings, data="photometry")

    # Multiclass
    # Plotting Confusion Matrix for just one seed
    if plots:
        print(
            lu.str_to_yellowstr(
                "Plotting confusion matrix for multiclass classification (Figure 4.)"
            )
        )
        for target in [3, 7]:
            settings.prediction_files = [
                settings.models_dir
                + "/"
                + model.strip("DES_").replace("CLF_2", f"S_0_CLF_{target}")
                + "/"
                + model.replace("DES_", "PRED_DES_").replace(
                    "CLF_2", f"S_0_CLF_{target}"
                )
                + ".pickle"
                for model in [Base, Var, BBB]
            ]
            if debug:
                print(settings.prediction_files)
            else:
                sp.science_plots(settings, onlycnf=True)
        # Uncomment to see some examples of seven-way classification
        # print(lu.str_to_yellowstr("Plotting candidates for multiclass classification"))
        # model_file = f"{settings.models_dir}/{Base.replace('DES_vanilla_','vanilla_S_0_').replace('CLF_2','CLF_7')}/{Base.replace('DES_vanilla_','vanilla_S_0_').replace('CLF_2','CLF_7')}.pt"
        # if os.path.exists(model_file):
        #     model_settings = conf.get_settings_from_dump(model_file)
        #     early_prediction.make_early_prediction(model_settings, nb_lcs=20)
        # else:
        #     print(lu.str_to_redstr(f"File not found {model_file}"))


def bayesian(df, df_delta, df_delta_ood, settings, plots, debug):
    """
    Bayesian RNNs: BBB and Variational
    """

    # 2. Variational hyper-parameters
    sel_criteria = ["DES_variational_CLF_2_R_None_saltfit_DF_0.2_N_global_lstm_32x2_"]
    print(lu.str_to_bluestr(f"Hyperparameters {sel_criteria}"))
    if not debug:
        sm.get_metric_ranges(df, sel_criteria)

    # 3. BBB hyper-parameters
    sel_criteria = ["DES_bayesian_CLF_2_R_None_saltfit_DF_0.2_N_global_lstm_32x2_"]
    print(lu.str_to_bluestr(f"Hyperparameters {sel_criteria}"))
    if not debug:
        sm.get_metric_ranges(df, sel_criteria)

    # 2 and 3. Best models
    print(lu.str_to_bluestr("Best performing Bayesian accuracies"))
    if debug:
        print(
            [
                Var,
                BBB,
                Var.replace("None", "zpho"),
                BBB.replace("None", "zpho"),
                Var.replace("None", "zspe"),
                BBB.replace("None", "zspe"),
            ]
        )
    else:
        sm.acc_auc_df(
            df,
            [
                Var,
                BBB,
                Var.replace("None", "zpho"),
                BBB.replace("None", "zpho"),
                Var.replace("None", "zspe"),
                BBB.replace("None", "zspe"),
            ],
        )
    # contamination
    # baseline saltfit, 1.0, all redshifts
    for model in [Var, BBB]:
        sel_criteria = model.split("None")
        if debug:
            print("contamination")
            print(sel_criteria)
        else:
            sm.print_contamination(df, sel_criteria, settings, data="photometry")

    # 4. Uncertainties
    print(lu.str_to_bluestr("Best performing Bayesian uncertainties"))
    print("Epistemic behaviour")
    for model in [Var, BBB]:
        m_right = model.replace("photometry", "saltfit")
        m_left = model.replace("photometry", "saltfit").replace("DF_1.0", "DF_0.5")
        print("salt", m_left, m_right)
        df_sel = df_delta[
            (df_delta["model_name_left"] == m_left)
            & (df_delta["model_name_right"] == m_right)
        ]
        if not debug:
            sm.nice_df_print(
                df_sel,
                keys=[
                    "all_accuracy_mean_delta",
                    "mean_all_class0_std_dev_mean_delta",
                    "all_entropy_mean_delta",
                ],
            )
        m_right = model
        m_left = model.replace("DF_1.0", "DF_0.43")
        print("complete", m_left, m_right)
        df_sel = df_delta[
            (df_delta["model_name_left"] == m_left)
            & (df_delta["model_name_right"] == m_right)
        ]
        if not debug:
            sm.nice_df_print(
                df_sel,
                keys=[
                    "all_accuracy_mean_delta",
                    "mean_all_class0_std_dev_mean_delta",
                    "all_entropy_mean_delta",
                ],
            )

    print("Uncertainty size")
    df_sel = df[df["model_name_noseed"].isin([Var, BBB])]
    df_sel = df_sel.round(4)
    if not debug:
        sm.nice_df_print(
            df_sel, keys=["mean_all_class0_std_dev_mean", "mean_all_class0_std_dev_std"]
        )

    if plots:
        print(
            lu.str_to_yellowstr(
                "Plotting candidates for multiclass classification (Fig. 5)"
            )
        )
        for model in [Var, BBB]:
            model_file = (
                f"{settings.models_dir}/"
                + model.replace("DES_", "").replace("CLF_2", "S_0_CLF_7")
                + "/"
                + model.replace("DES_", "").replace("CLF_2", "S_0_CLF_7")
                + ".pt"
            )
            if os.path.exists(model_file):
                if debug:
                    print(model_file)
                else:
                    model_settings = conf.get_settings_from_dump(model_file)
                    early_prediction.make_early_prediction(
                        model_settings, nb_lcs=20, do_gifs=True
                    )
            else:
                print(lu.str_to_redstr(f"File not found {model_file}"))
        print(lu.str_to_yellowstr("Adding gifs for binary classification"))
        for model in [Var, BBB]:
            model_file = (
                f"{settings.models_dir}/"
                + model.strip("DES_").replace("CLF_2", "S_0_CLF_2")
                + "/"
                + model.strip("DES_").replace("CLF_2", "S_0_CLF_2")
                + ".pt"
            )
            if os.path.exists(model_file):
                if debug:
                    print(model_file)
                else:
                    model_settings = conf.get_settings_from_dump(model_file)
                    early_prediction.make_early_prediction(
                        model_settings, nb_lcs=10, do_gifs=True
                    )


def towards_cosmo(df, df_delta, df_delta_ood, settings, plots, debug):
    """
    Towards cosmology
    """
    # 1. Calibration
    print(lu.str_to_bluestr("Calibration"))
    df_sel = df.copy()
    df_sel = df_sel.round(4)
    # rf can't be done with photometry
    df_sel = df_sel[
        (
            df_sel["model_name_noseed"].isin(
                [item.replace("photometry", "saltfit") for item in list_models]
            )
        )
        & (df_sel["source_data"] == "saltfit")
    ]
    print("saltfit")
    if not debug:
        sm.nice_df_print(
            df_sel,
            keys=[
                "model_name_noseed",
                "calibration_dispersion_mean",
                "calibration_dispersion_std",
            ],
        )
    # without rf
    print("photometry")
    df_sel = df.copy()
    df_sel = df_sel[
        (df_sel["model_name_noseed"].isin([Base, Var, BBB]))
        & (df_sel["source_data"] == "photometry")
    ]
    if not debug:
        sm.nice_df_print(
            df_sel,
            keys=[
                "model_name_noseed",
                "calibration_dispersion_mean",
                "calibration_dispersion_std",
            ],
        )
    # Calibration vs. training set size
    # using salt
    print(lu.str_to_bluestr("Calibration vs. data set size"))
    print("Baseline")
    sel_criteria = Base_salt.split("DF_1.0")
    if debug:
        print(sel_criteria)
    else:
        sm.get_metric_ranges(
            df, sel_criteria, metric="calibration_dispersion", round_output=5
        )
    # Calibration vs. dataset nature
    sel_criteria = [
        "DES_vanilla_CLF_2_R_None_saltfit_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C"
    ]
    if debug:
        print(sel_criteria)
    else:
        sm.get_metric_ranges(
            df, sel_criteria, metric="calibration_dispersion", round_output=5
        )
    # Calibration figure
    if plots:
        print(lu.str_to_yellowstr("Plotting reliability diagram (Figure 6)"))
        tmp_pred_files = settings.prediction_files
        settings.prediction_files = [
            settings.models_dir
            + "/"
            + model.strip("DES_")
            .replace("CLF_2", "S_0_CLF_2")
            .replace("photometry", "saltfit")
            + "/"
            + model.replace("DES_", "PRED_DES_")
            .replace("CLF_2", "S_0_CLF_2")
            .replace("photometry", "saltfit")
            + ".pickle"
            for model in [RF, Base, Var, BBB]
        ]
        if debug:
            print(settings.prediction_files)
        else:
            sp.plot_calibration(settings)
        settings.prediction_files = tmp_pred_files

    # 2. Representativeness
    print(lu.str_to_bluestr("Representativeness"))
    for model in [Base, Var, BBB]:
        m_left = model.replace("photometry", "saltfit")
        m_right = model.replace("DF_1.0", "DF_0.43")
        print(m_left, m_right)
        df_sel = df_delta[
            (df_delta["model_name_left"] == m_left)
            & (df_delta["model_name_right"] == m_right)
        ]
        if not debug:
            sm.nice_df_print(
                df_sel,
                keys=[
                    "all_accuracy_mean_delta",
                    "mean_all_class0_std_dev_mean_delta",
                    "all_entropy_mean_delta",
                ],
            )

    # 3. OOD
    print(lu.str_to_bluestr("Out-of-distribution light-curves"))
    # OOD type assignement figure
    if plots:
        print(lu.str_to_yellowstr("Plotting OOD classification percentages (Figure 8)"))
        if not debug:
            sp.create_OOD_classification_plots(df, list_models_rnn, settings)
    # Get entropy
    print("binary")
    df_sel = df_delta_ood[df_delta_ood["model_name_noseed"].isin(list_models_rnn)]
    if not debug:
        sm.nice_df_print(df_sel)
    print("ternary")
    list_models_sel = [item.replace("CLF_2", "CLF_3") for item in list_models_rnn]
    df_sel = df_delta_ood[df_delta_ood["model_name_noseed"].isin(list_models_sel)]
    if not debug:
        sm.nice_df_print(df_sel)
    print("seven-way")
    list_models_sel = [item.replace("CLF_2", "CLF_7") for item in list_models_rnn]
    df_sel = df_delta_ood[df_delta_ood["model_name_noseed"].isin(list_models_sel)]
    if not debug:
        sm.nice_df_print(df_sel)
    if plots:
        print(
            lu.str_to_yellowstr(
                "Plotting OOD candidates with seven-way classification (Figure 9)"
            )
        )
        model_file = (
            f"{settings.models_dir}/"
            + Var.replace("DES_variational_", "variational_S_0_").replace(
                "CLF_2", "CLF_7"
            )
            + "/"
            + Var.replace("DES_variational_", "variational_S_0_").replace(
                "CLF_2", "CLF_7"
            )
            + ".pt"
        )
        if os.path.exists(model_file):
            if debug:
                print(model_file)
            else:
                model_settings = conf.get_settings_from_dump(model_file)
                early_prediction.make_early_prediction(model_settings, nb_lcs=20)
        else:
            print(lu.str_to_redstr(f"File not found {model_file}"))

    # 4. Cosmology
    print(lu.str_to_bluestr("SNe Ia for cosmology"))
    # Plotting Hubble residuals adn other science plots for just one seed
    if plots:
        print(lu.str_to_yellowstr("Plotting Hubble residuals (Figures 10 and 11)"))
        tmp_pred_files = settings.prediction_files
        settings.prediction_files = [
            settings.models_dir
            + "/"
            + model.strip("DES_").replace("CLF_2", "S_0_CLF_2")
            + "/"
            + model.replace("DES_", "PRED_DES_").replace("CLF_2", "S_0_CLF_2")
            + ".pickle"
            for model in [Base, Var, BBB]
        ]
        if debug:
            print(settings.prediction_files)
        else:
            sp.science_plots(settings)
        settings.prediction_files = tmp_pred_files



import os
import warnings
import math
import numpy as np
import pandas as pd
from tqdm import tqdm
from time import time
from pathlib import Path
from sklearn import metrics
import torch
import torch.nn as nn

from ..utils import data_utils as du
from ..utils import training_utils as tu
from ..utils import logging_utils as lu
from supernnova.utils.swag_utils import SwagModel


def find_idx(array, value):
    """Utility to find the index of the element of ``array`` that most closely
    matches ``value``

    Args:
        array (np.array): The array in which to search
        value (float): The value for which we are looking for a match

    Returns:
        (int) the index of of the element of ``array`` that most closely
        matches ``value``

    """

    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (
        idx == len(array)
        or math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])
    ):
        return idx - 1
    else:
        return idx


def get_batch_predictions(rnn, X, target):
    """Utility to obtain predictions for a given batch

    Args:
        rnn (torch.nn): The RNN model
        X (torch.Tensor): The batch on which to carry out predictions
        target (torch.longTensor): The true class of each element in the batch

    Returns:
        Tuple containing

        - arr_preds (np.array): predictions
        - arr_target (np.array): actual targets

    """

    out = rnn.forward(X)
    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
    arr_target = target.detach().cpu().numpy()

    return arr_preds, arr_target


def get_batch_predictions_MFE(rnn, X, target):
    """Utility to obtain predictions for a given batch

    Args:
        rnn (torch.nn): The RNN model
        X (torch.Tensor): The batch on which to carry out predictions
        target (torch.longTensor): The true class of each element in the batch

    Returns:
        Tuple containing

        - arr_preds (np.array): predictions
        - arr_target (np.array): actual targets

    """

    out = rnn.forward(X, mean_field_inference=True)
    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
    arr_target = target.detach().cpu().numpy()

    return arr_preds, arr_target


def get_batch_predictions_SWAG(model: SwagModel, X, target, scale: float, cov=True):
    """Utility to obtain predictions for a given batch

    Args:
        model: The SwagModel model
        X (torch.Tensor): The batch on which to carry out predictions
        target (torch.longTensor): The true class of each element in the batch
        scale (float): The scale parameter for covariance
        cov (bool): If True, enable calculating low-rank covariance

    Returns:
        Tuple containing

        - arr_preds (np.array): predictions
        - arr_target (np.array): actual targets

    """

    sample_model = model.sample(scale, cov)
    out = sample_model.forward(X)
    arr_preds = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
    arr_target = target.detach().cpu().numpy()

    return arr_preds, arr_target


def dispatch_batch_predictions(model, X, target, mode=None, **kwargs):
    """
    Dispatch different batch prediction functions based on given mode.
    """
    if mode is None:
        arr_preds, arr_target = get_batch_predictions(model, X, target)
    elif mode == "MFE":
        arr_preds, arr_target = get_batch_predictions_MFE(model, X, target)
    elif mode == "SWAG":
        scale = kwargs.get("scale", 0.5)
        cov = kwargs.get("cov", True)
        arr_preds, arr_target = get_batch_predictions_SWAG(model, X, target, scale, cov)
    else:
        raise ValueError(f"Invalid mode: {mode}")
    return arr_preds, arr_target


def construct_predict_dict(n_elements, n_samples, nb_classes, MFE=False):
    """Construct a dict to hold prediction results."""
    d_pred = {
        key: np.zeros((n_elements, n_samples, nb_classes)).astype(np.float32)
        for key in ["all"]
        + [f"PEAKMJD{offset}" for offset in du.OFFSETS_STR if not MFE]
        + [f"all_{OOD}" for OOD in du.OOD_TYPES]
    }

    d_pred["target"] = np.zeros((n_elements, n_samples)).astype(np.int64)
    d_pred["SNID"] = np.zeros((n_elements, n_samples)).astype(str)

    return d_pred


def flatten_to_dataframe(pred_dict):
    """Flatten the prediction dict and convert it to pandas.DataFrame."""
    d_series = {}
    (n_elements, n_samples, nb_classes) = pred_dict["all"].shape
    for (key, value) in pred_dict.items():
        value = value.reshape((n_elements * n_samples, -1))
        value_dim = value.shape[1]
        if value_dim == 1:
            d_series[key] = np.ravel(value)
        else:
            for i in range(value_dim):
                d_series[f"{key}_class{i}"] = value[:, i]

    return pd.DataFrame.from_dict(d_series)


def get_aggregated_dataframe(pred_df):
    med_pred = pred_df.groupby("SNID").median()
    med_pred.columns = [str(col) + "_median" for col in med_pred.columns]
    std_pred = pred_df.groupby("SNID").std()
    std_pred.columns = [str(col) + "_std" for col in std_pred.columns]
    df_agg = pd.merge(med_pred, std_pred, on="SNID")
    df_agg["SNID"] = df_agg.index
    df_agg["target"] = df_agg["target_median"]

    return df_agg


def get_full_accuracy(pred_df, settings, model):
    g_pred = pred_df.groupby("SNID").median()
    preds = g_pred[[f"all_class{i}" for i in range(settings.nb_classes)]].values
    preds = np.argmax(preds, 1)
    acc = (preds == g_pred.target.values).sum() / len(g_pred)

    lu.print_green(f"Full Accuracy {model}", acc)
    for col in [f"PEAKMJD{s}" for s in du.OFFSETS_STR]:

        preds_target = g_pred[
            [f"{col}_class{i}" for i in range(settings.nb_classes)] + ["target"]
        ].dropna()
        preds = preds_target[
            [f"{col}_class{i}" for i in range(settings.nb_classes)]
        ].values
        target = preds_target["target"].values
        preds = np.argmax(preds, 1)
        acc = (preds == target).sum() / len(g_pred)

        lu.print_green(f"{col} Accuracy", acc)

    print()
    print()


def get_other_accuracy(pred_df, settings, mode):
    class_col = [f"all_class{i}" for i in range(settings.nb_classes)]
    tmp = pred_df[["SNID", "target"] + class_col].groupby("SNID").mean()
    preds = np.argmax(tmp[class_col].values, 1)
    acc = (preds == tmp.target.values).sum() / len(tmp)
    lu.print_green(f"Accuracy {mode}", acc)
    lu.print_green(
        "Balanced Accuracy",
        metrics.balanced_accuracy_score(tmp.target.values, preds),
    )

    for OOD in ["random", "reverse", "shuffle", "sin"]:
        class_col_ood = [f"all_{OOD}_class{i}" for i in range(settings.nb_classes)]
        entropy_ood = (
            -(pred_df[class_col_ood].values * np.log(pred_df[class_col_ood].values))
            .sum(1)
            .mean()
        )
        entropy = (
            -(pred_df[class_col].values * np.log(pred_df[class_col].values))
            .sum(1)
            .mean()
        )
        lu.print_green(f"Delta Entropy {OOD} {mode}", entropy_ood - entropy)

    print()
    print()


def get_predictions(settings, model_file=None):
    """Obtain predictions for a given RNN model specified by the
    ``settings`` argument or alternatively, by a model_file

    - Models are benchmarked on the test data set
    - Batch size can be controled to speed up predictions
    - For Bayesian models, multiple predictions are carried to
        obtain a distribution of predictions
    - Predictions are computed for full lightcurves, and around the peak light
    - Predictions are saved to a pickle file (for faster loading)

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        model_file (str): Path to saved model weights. Default: ``None``

    Returns:
        out_files (list): a list of prediction files
    """

    settings.random_length = False
    settings.random_redshift = False

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # set the default behaviour: get prediction use RNN model only
    no_rnn = 0
    no_swag = 1

    # set output file
    out_files = []

    # Load model files
    if model_file is None:
        dump_dir = f"{settings.models_dir}/{settings.pytorch_model_name}"
        model_file = f"{dump_dir}/{settings.pytorch_model_name}.pt"

        # Load SWAG model
        if settings.swag:
            no_swag = 0
            swag_model_file = f"{dump_dir}/{settings.pytorch_model_name}_swag.pt"
            if not os.path.exists(swag_model_file):
                warnings.warn("SWAG model does not exist.")
                no_swag = 1
    else:
        dump_dir = f"{settings.dump_dir}/models/{settings.pytorch_model_name}"
        os.makedirs(dump_dir, exist_ok=True)
        if model_file.endswith("_swag.pt"):
            swag_model_file = model_file
            no_rnn = 1
            no_swag = 0

    # Load the data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Batching stuff together
    num_elem = len(list_data_test)
    num_batches = num_elem / min(num_elem, settings.batch_size)
    list_batches = np.array_split(np.arange(num_elem), num_batches)

    # Fetch SN info
    df_SNinfo = du.load_HDF5_SNinfo(settings).set_index("SNID")

    def _populate_predict_dict(_model, _pred_dict, _n_samples, _mode=None, **kwargs):
        """Get prediction results and store them in the given dict."""
        # Loop over data and make prediction
        for batch_idxs in tqdm(
            list_batches, desc="Computing predictions on test set", ncols=100
        ):

            start_idx, end_idx = batch_idxs[0], batch_idxs[-1] + 1
            SNIDs = [data[2] for data in list_data_test[start_idx:end_idx]]

            peak_MJDs = df_SNinfo.loc[SNIDs]["PEAKMJDNORM"].values
            delta_times = [
                data[3][:, settings.d_feat_to_idx["delta_time"]]
                for data in list_data_test[start_idx:end_idx]
            ]
            times = [np.cumsum(t) for t in delta_times]

            with torch.no_grad():

                #############################
                # Full lightcurve prediction
                #############################

                packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                    list_data_test, batch_idxs, settings
                )

                for iter_ in tqdm(range(_n_samples), ncols=100):

                    arr_preds, arr_target = dispatch_batch_predictions(
                        _model, packed, target_tensor, mode=_mode, **kwargs
                    )

                    # Rever sorting that occurs in get_batch_predictions
                    arr_preds = arr_preds[idxs_rev_sort]
                    arr_target = arr_target[idxs_rev_sort]

                    _pred_dict["all"][start_idx:end_idx, iter_] = arr_preds
                    _pred_dict["target"][start_idx:end_idx, iter_] = arr_target
                    _pred_dict["SNID"][start_idx:end_idx, iter_] = SNIDs

                if _mode != "MFE":
                    #############################
                    # Predictions around PEAKMJD
                    #############################
                    for offset in du.OFFSETS_VAL:
                        slice_idxs = [
                            find_idx(times[k], peak_MJDs[k] + offset)
                            for k in range(len(times))
                        ]
                        # Split in 2 arrays:
                        # oob_idxs: the slice for early prediction is empty for those indices
                        # inb_idxs: the slice is not empty
                        oob_idxs = np.where(np.array(slice_idxs) < 1)[0]
                        inb_idxs = np.where(np.array(slice_idxs) >= 1)[0]

                        if len(inb_idxs) > 0:
                            # We only carry out prediction for samples in ``inb_idxs``
                            offset_batch_idxs = [batch_idxs[b] for b in inb_idxs]
                            max_lengths = [slice_idxs[b] for b in inb_idxs]
                            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                                list_data_test,
                                offset_batch_idxs,
                                settings,
                                max_lengths=max_lengths,
                            )

                            for iter_ in tqdm(range(_n_samples), ncols=100):

                                arr_preds, arr_target = dispatch_batch_predictions(
                                    _model, packed, target_tensor, mode=_mode, **kwargs
                                )

                                # Rever sorting that occurs in get_batch_predictions
                                arr_preds = arr_preds[idxs_rev_sort]

                                suffix = str(offset) if offset != 0 else ""
                                suffix = f"+{suffix}" if offset > 0 else suffix
                                col = f"PEAKMJD{suffix}"

                                _pred_dict[col][start_idx + inb_idxs, iter_] = arr_preds
                                # For oob_idxs, no prediction can be made, fill with nan
                                _pred_dict[col][start_idx + oob_idxs, iter_] = np.nan

                #############################
                # OOD predictions
                #############################

                for OOD in ["random", "shuffle", "reverse", "sin"]:
                    packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                        list_data_test, batch_idxs, settings, OOD=OOD
                    )

                    for iter_ in tqdm(range(_n_samples), ncols=100):

                        arr_preds, arr_target = dispatch_batch_predictions(
                            _model, packed, target_tensor, mode=_mode, **kwargs
                        )
                        # Revert sorting that occurs in get_batch_predictions
                        arr_preds = arr_preds[idxs_rev_sort]
                        arr_target = arr_target[idxs_rev_sort]

                        _pred_dict[f"all_{OOD}"][start_idx:end_idx, iter_] = arr_preds

    if no_rnn == 0:
        lu.print_green("Processing prediction using RNN model")
        # Load RNN model
        rnn = tu.get_model(settings, len(settings.training_features))
        rnn_state = torch.load(
            model_file,
            map_location=lambda storage, loc: storage,
            weights_only=False,
        )
        rnn.load_state_dict(rnn_state)
        rnn.to(settings.device)
        rnn.eval()

        # Specify output file name
        prediction_file = f"{dump_dir}/PRED_{settings.pytorch_model_name}.pickle"

        out_files.append(prediction_file)

        # Prepare output dicts
        d_pred = construct_predict_dict(
            num_elem, settings.num_inference_samples, settings.nb_classes
        )
        d_pred_MFE = construct_predict_dict(num_elem, 1, settings.nb_classes, MFE=True)

        # Fill in the dict
        _populate_predict_dict(rnn, d_pred, settings.num_inference_samples)
        _populate_predict_dict(rnn, d_pred_MFE, 1, _mode="MFE")

        # Flatten all arrays and aggregate in dataframe
        df_pred = flatten_to_dataframe(d_pred)
        df_pred_MFE = flatten_to_dataframe(d_pred_MFE)

        # Save predictions
        df_pred.to_pickle(prediction_file, protocol=4)

        # Saving aggregated preds for bayesian models
        if settings.model == "variational" or settings.model == "bayesian":
            df_bayes = get_aggregated_dataframe(df_pred)
            bay_pred_file = prediction_file.replace(".pickle", "_aggregated.pickle")
            df_bayes.to_pickle(bay_pred_file, protocol=4)

        # Display accuracy
        get_full_accuracy(df_pred, settings, "RNN")
        get_other_accuracy(df_pred, settings, "MC")
        get_other_accuracy(df_pred_MFE, settings, "MFE")

        lu.print_green("Finished getting RNN predictions ")

    if no_swag == 0:
        lu.print_green("Processing prediction using SWA/SWAG model")

        # Check setting for cov
        if settings.swag_no_lr_cov:
            swag_cov = False
            cov_str = "no_lr_cov"
        else:
            swag_cov = True
            cov_str = "cov"

        # Load SWAG model
        swag_rnn = torch.load(swag_model_file, weights_only=False)
        swag_rnn.to(settings.device)

        # Specify output file names
        prediction_file_swa = (
            f"{dump_dir}/PRED_{settings.pytorch_model_name}_swa.pickle"
        )
        out_files.append(prediction_file_swa)

        prediction_file_swag = f"{dump_dir}/PRED_{settings.pytorch_model_name}_nsample_{settings.swag_samples}_scale_{settings.swag_scale}_{cov_str}_swag.pickle"
        out_files.append(prediction_file_swag)

        # Prepare output dicts
        d_pred_SWA = construct_predict_dict(num_elem, 1, settings.nb_classes)
        d_pred_SWAG = construct_predict_dict(
            num_elem, settings.swag_samples, settings.nb_classes
        )

        # Fill in the dict
        _populate_predict_dict(
            swag_rnn, d_pred_SWA, 1, _mode="SWAG", scale=0, cov=False
        )
        _populate_predict_dict(
            swag_rnn,
            d_pred_SWAG,
            settings.swag_samples,
            _mode="SWAG",
            scale=settings.swag_scale,
            cov=swag_cov,
        )

        # Flatten all arrays and aggregate in dataframe
        df_pred_SWA = flatten_to_dataframe(d_pred_SWA)
        df_pred_SWAG = flatten_to_dataframe(d_pred_SWAG)

        # Save predictions
        df_pred_SWA.to_pickle(prediction_file_swa, protocol=4)
        df_pred_SWAG.to_pickle(prediction_file_swag, protocol=4)

        # Saving aggregated preds for swag model
        df_bayes = get_aggregated_dataframe(df_pred_SWAG)
        bay_pred_file = prediction_file_swag.replace(".pickle", "_aggregated.pickle")
        df_bayes.to_pickle(bay_pred_file, protocol=4)

        # Display accuracy for SWA model
        get_full_accuracy(df_pred_SWA, settings, "SWA")
        get_other_accuracy(df_pred_SWA, settings, "SWA MC")

        # Display accuracy for SWAG model
        get_full_accuracy(df_pred_SWAG, settings, "SWAG")
        get_other_accuracy(df_pred_SWAG, settings, "SWAG MC")

        lu.print_green("Finished getting SWA/SWAG predictions ")

    return out_files


def get_predictions_for_speed_benchmark(settings):
    """Test RNN models inference speed

    - Models are benchmarked on the test data set
    - Batch size can be controled to speed up predictions
    - For Bayesian models, multiple predictions are carried to
        obtain a distribution of predictions
    - Results are saved to a .csv for future use

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    settings.random_length = False

    if "vanilla" in settings.pytorch_model_name:
        settings.num_inference_samples = 1

    # Load the data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Load RNN model
    rnn = tu.get_model(settings, len(settings.training_features))
    rnn.to(settings.device)
    rnn.eval()

    # Batching lightcurves
    num_elem = len(list_data_test)
    num_batches = num_elem / min(num_elem, settings.batch_size)
    list_batches = np.array_split(np.arange(num_elem), num_batches)

    start_time = time()

    # Loop over data and predict
    for batch_idxs in tqdm(list_batches, ncols=100):

        with torch.no_grad():

            #############################
            # Full lightcurve prediction
            #############################

            packed, _, target_tensor, _ = tu.get_data_batch(
                list_data_test, batch_idxs, settings
            )

            for iter_ in tqdm(range(settings.num_inference_samples), ncols=100):

                arr_preds, arr_target = get_batch_predictions(
                    rnn, packed, target_tensor
                )

    total_time = time() - start_time
    supernova_per_s = num_elem / total_time
    model_id = f"{settings.model}_{settings.batch_size}_{settings.device}"

    df = pd.DataFrame(
        data=np.array([supernova_per_s]).astype(np.float16), columns=["Supernova_per_s"]
    )
    df["model"] = settings.model
    df["batch_size"] = settings.batch_size
    df["device"] = settings.device
    df["id"] = model_id

    # Save results to csv
    results_file = Path(settings.stats_dir) / "rnn_speed.csv"
    try:
        df_all = pd.read_csv(results_file, index_col="id")
        df_all.loc[model_id] = df.set_index("id").loc[model_id]
        df_all.reset_index().to_csv(results_file, index=False)
    except Exception:
        df.to_csv(results_file, index=False)



import math
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F


def log_norm(x, min_clip, mean, std, F=torch):
    """ """

    x = (F.log(x - min_clip + 1e-5) - mean) / std
    return x


def inverse_log_norm(x, min_clip, mean, std, F=torch):

    x = F.exp(x * std + mean) + min_clip - 1e-5

    return x


class BayesianRNN(torch.nn.Module):
    def __init__(self, input_size, settings):
        super().__init__()

        # Params
        self.layer_type = settings.layer_type
        self.hidden_size = settings.hidden_dim
        self.output_size = settings.nb_classes
        self.num_layers = settings.num_layers
        self.bidirectional = settings.bidirectional
        self.use_cuda = settings.use_cuda
        self.rnn_output_option = settings.rnn_output_option

        self.prior = Prior(settings.pi, settings.log_sigma1, settings.log_sigma2)
        self.prior_output = Prior(
            settings.pi, settings.log_sigma1_output, settings.log_sigma2_output
        )

        bidirectional_factor = 2 if self.bidirectional is True else 1
        last_input_size = (
            self.hidden_size * bidirectional_factor
            if self.rnn_output_option == "mean"
            else self.hidden_size * bidirectional_factor * self.num_layers
        )

        init_recurrent = {
            "mu_lower": -0.05,
            "mu_upper": 0.05,
            "rho_lower": math.log(
                math.exp(self.prior.sigma_mix / settings.rho_scale_lower) - 1.0
            ),
            "rho_upper": math.log(
                math.exp(self.prior.sigma_mix / settings.rho_scale_upper) - 1.0
            ),
        }

        init = {
            "mu_lower": -0.05,
            "mu_upper": 0.05,
            "rho_lower": math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_lower_output)
                - 1.0
            ),
            "rho_upper": math.log(
                math.exp(self.prior_output.sigma_mix / settings.rho_scale_upper_output)
                - 1.0
            ),
        }

        # Define layers
        self.rnn_layer = BayesLSTM(
            input_size,
            self.hidden_size,
            num_layers=self.num_layers,
            dropout=0,
            bidirectional=self.bidirectional,
            batch_first=False,
            bias=True,
            prior=self.prior,
            **init_recurrent,
        )
        self.output_layer = BayesLinear(
            last_input_size, self.output_size, self.prior_output, **init
        )

        self.kl = None

    def forward(self, x, mean_field_inference=False):
        # out = packed output from last layer
        # out has dim (seq_len, batch_size, hidden_size) when unpacked
        # hidden = (hn, cn) for lstm
        # hidden = hn for GRU and RNN
        # hn has dim (num_layers * num_directions, batch, hidden_size)
        # cn has dim (num_layers * num_directions, batch, hidden_size)
        # assuming num_directions = 1, num_layers = 2 :
        # hn[-1, -1] == out[len, -1] where len is the len of the seq at batch index == -1

        x, hidden = self.rnn_layer(x, mean_field_inference=mean_field_inference)

        # Output options
        # Standard: all layers, only end of pass
        #    - take last pass in all layers (hidden)
        #    - reshape and apply dropout
        #    - use h20 to obtain output (h2o input: hidden_size*num_layers*bi)
        # Mean: last layer, mean on sequence
        #    - take packed output from last layer (out) that contains all time steps for the last layer
        #    - find where padding was done and create a mask for those values, apply this mask
        #    - take a mean for the whole sequence (time_steps)
        #    - use h2o to obtain output (beware! it is only one layer deep since it is the last one only)

        if self.rnn_output_option == "standard":
            # Special case for lstm where hidden = (h, c)
            if self.layer_type == "lstm":
                hn = hidden[0]
            else:
                hn = hidden

            hn = hn.permute(1, 2, 0).contiguous()
            # hn is (num_layers * num_directions, batch, hidden_size)
            batch_size = hn.shape[0]
            # hn now is (batch, hidden size, num_layers * num_directions)
            x = hn.view(batch_size, -1)
            # x is (batch, hidden size * num_layers * num_directions)

        if self.rnn_output_option == "mean":
            if isinstance(x, torch.nn.utils.rnn.PackedSequence):
                x, lens = torch.nn.utils.rnn.pad_packed_sequence(x)
                # x is (seq_len, batch, hidden size)

                # take mean over seq_len
                x = x.sum(0) / lens.unsqueeze(-1).float().to(x.device)
                # x is (batch, hidden_size)
            else:
                x = x.mean(0)
                # x is (batch, hidden_size)

        # Final projection layer
        output = self.output_layer(x, mean_field_inference=mean_field_inference)

        # Compute KL
        self.kl = self.rnn_layer.kl + self.output_layer.kl

        return output


class BayesLSTM(nn.Module):
    def __init__(
        self,
        input_size,
        hidden_size,
        num_layers=1,
        bias=True,
        batch_first=False,
        dropout=0.0,
        bidirectional=False,
        prior=None,
        mu_lower=-0.05,
        mu_upper=0.05,
        rho_lower=math.log(math.exp(1.0 / 4.0) - 1.0),
        rho_upper=math.log(math.exp(1.0 / 2.0) - 1.0),
    ):

        super().__init__()
        self.module = nn.LSTM(
            input_size,
            hidden_size,
            num_layers,
            bias,
            batch_first,
            dropout,
            bidirectional,
        )

        self.prior = prior
        self.kl = None
        for name, w in self.module.named_parameters():
            mu, rho = get_bbb_variable(
                w.shape, mu_lower, mu_upper, rho_lower, rho_upper
            )
            self.register_parameter(f"{name}_mu", mu)
            self.register_parameter(f"{name}_rho", rho)

            self.module._parameters[name] = mu

    def _setweights(self, mean_field_inference=False):
        kl = 0
        for name, _ in self.module.named_parameters():
            mu = getattr(self, f"{name}_mu")
            rho = getattr(self, f"{name}_rho")

            sigma = F.softplus(rho) + 1e-5

            if mean_field_inference is False:
                eps = rho.data.new(rho.size()).normal_(0.0, 1.0)
                w = mu + sigma * eps
                kl += compute_KL(w, mu, sigma, self.prior)
            else:
                w = mu

            self.module._parameters[name] = w

        self.kl = kl

    def forward(self, *args, mean_field_inference=False):
        self._setweights()
        self.module.flatten_parameters()
        with warnings.catch_warnings():
            # To avoid the warning that comes because the weights aren't flattened.
            warnings.simplefilter("ignore")
            return self.module.forward(*args)


class BayesLinear(nn.Module):
    def __init__(
        self, in_features, out_features, prior, mu_lower, mu_upper, rho_lower, rho_upper
    ):
        super(BayesLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.prior = prior

        mu, rho = get_bbb_variable(
            (out_features, in_features), mu_lower, mu_upper, rho_lower, rho_upper
        )

        bias = nn.Parameter(torch.Tensor(out_features))
        bias.data.fill_(0.0)

        self.mu = mu
        self.rho = rho
        self.bias = bias
        self.kl = None

    def forward(self, x, mean_field_inference=False):

        self.kl = 0

        mu = self.mu

        if mean_field_inference is False:
            sigma = F.softplus(self.rho) + 1e-5
            eps = mu.data.new(mu.size()).normal_(0.0, 1.0)
            weights = mu + eps * sigma
            # Compute KL divergence
            self.kl = compute_KL(weights, mu, sigma, self.prior)
        else:
            weights = mu

        logits = F.linear(x, weights, self.bias)

        return logits

    def __repr__(self):
        return (
            self.__class__.__name__
            + " ("
            + str(self.in_features)
            + " -> "
            + str(self.out_features)
            + ")"
        )


class Prior(object):
    def __init__(self, pi=0.25, log_sigma1=-1.0, log_sigma2=-7.0):
        self.pi_mixture = pi
        self.log_sigma1 = log_sigma1
        self.log_sigma2 = log_sigma2
        self.sigma1 = math.exp(log_sigma1)
        self.sigma2 = math.exp(log_sigma2)

        self.sigma_mix = math.sqrt(
            pi * math.pow(self.sigma1, 2) + (1.0 - pi) * math.pow(self.sigma2, 2)
        )


def get_bbb_variable(shape, mu_lower, mu_upper, rho_lower, rho_upper):

    mu = nn.Parameter(torch.Tensor(*shape))
    rho = nn.Parameter(torch.Tensor(*shape))

    # Initialize
    mu.data.uniform_(mu_lower, mu_upper)
    rho.data.uniform_(rho_lower, rho_upper)

    return mu, rho


def compute_KL(x, mu, sigma, prior):

    posterior = torch.distributions.Normal(mu.view(-1), sigma.view(-1))
    log_posterior = posterior.log_prob(x.view(-1)).sum()

    if x.is_cuda:
        n1 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma1]).cuda()
        )
        n2 = torch.distributions.Normal(
            torch.tensor([0.0]).cuda(), torch.tensor([prior.sigma2]).cuda()
        )
    else:
        n1 = torch.distributions.Normal(0.0, prior.sigma1)
        n2 = torch.distributions.Normal(0.0, prior.sigma2)

    mix1 = torch.sum(n1.log_prob(x)) + math.log(prior.pi_mixture)
    mix2 = torch.sum(n2.log_prob(x)) + math.log(1.0 - prior.pi_mixture)
    prior_mix = torch.stack([mix1, mix2])
    log_prior = torch.logsumexp(prior_mix, 0)

    return log_posterior - log_prior



import torch
import json
import numpy as np
import torch.nn as nn
from tqdm import tqdm
from time import time
from pathlib import Path
from ..utils import training_utils as tu
from ..utils import logging_utils as lu
from ..utils.swag_utils import SwagModel


def get_lr(settings):
    """Select optimal starting learning rate when training with a 1-cycle policy

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    # Data
    list_data_train, list_data_val = tu.load_HDF5(settings, test=False)

    num_elem = len(list_data_train)
    num_batches = num_elem // min(num_elem // 2, settings.batch_size)
    list_batches = np.array_split(np.arange(num_elem), num_batches)
    np.random.shuffle(list_batches)

    lr_init_value = 1e-8
    lr = float(lr_init_value)
    lr_final_value = 10.0
    beta = 0.98
    avg_loss = 0.0
    best_loss = 0.0
    batch_num = 0
    list_losses = []
    list_lr = []
    mult = (lr_final_value / lr_init_value) ** (1 / num_batches)

    settings.learning_rate = lr_init_value

    # Model specification
    rnn = tu.get_model(settings, len(settings.training_features))
    criterion = nn.CrossEntropyLoss()
    optimizer = tu.get_optimizer(settings, rnn)

    # Prepare for GPU if required
    if settings.use_cuda:
        rnn.cuda()
        criterion.cuda()

    for batch_idxs in tqdm(list_batches, ncols=100):

        batch_num += 1

        # Sample a batch in packed sequence form
        packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
            list_data_train, batch_idxs, settings
        )
        # Train step : forward backward pass
        loss = tu.train_step(
            settings,
            rnn,
            packed,
            target_tensor,
            criterion,
            optimizer,
            target_tensor.size(0),
            len(list_batches),
        )
        loss = loss.detach().cpu().numpy().item()

        # Compute the smoothed loss
        avg_loss = beta * avg_loss + (1 - beta) * loss
        smoothed_loss = avg_loss / (1 - beta**batch_num)
        # Stop if the loss is exploding
        if batch_num > 1 and smoothed_loss > 4 * best_loss:
            break
        # Record the best loss
        if smoothed_loss < best_loss or batch_num == 1:
            best_loss = smoothed_loss
        # Store the values
        list_losses.append(smoothed_loss)
        list_lr.append(lr)
        # Update the lr for the next step
        lr *= mult

        # Set learning rate
        for param_group in optimizer.param_groups:

            param_group["lr"] = lr

    idx_min = np.argmin(list_losses)
    print("Min loss", list_losses[idx_min], "LR", list_lr[idx_min])

    return list_lr[idx_min]


def train_cyclic(settings):
    """Train RNN models with a 1-cycle policy

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """
    # save training data config
    save_normalizations(settings)

    max_learning_rate = get_lr(settings) / 10
    min_learning_rate = max_learning_rate / 10
    settings.learning_rate = min_learning_rate
    print("Setting learning rate to", min_learning_rate)

    def one_cycle_sched(epoch, minv, maxv, phases):
        if epoch <= phases[0]:
            out = minv + (maxv - minv) / (phases[0]) * epoch
        elif phases[0] < epoch <= phases[1]:
            increment = (minv - maxv) / (phases[1] - phases[0])
            out = maxv + increment * (epoch - phases[0])
        else:
            increment = (minv / 100 - minv) / (phases[2] - phases[1])
            out = minv + increment * (epoch - phases[1])

        return out

    # Data
    list_data_train, list_data_val = tu.load_HDF5(settings, test=False)

    # Model specification
    rnn = tu.get_model(settings, len(settings.training_features))
    criterion = nn.CrossEntropyLoss()
    optimizer = tu.get_optimizer(settings, rnn)

    # Prepare for GPU if required
    if settings.use_cuda:
        rnn.cuda()
        criterion.cuda()

    # Keep track of losses for plotting
    loss_str = ""
    d_monitor_train = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    d_monitor_val = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    if "bayesian" in settings.pytorch_model_name:
        d_monitor_train["KL"] = []
        d_monitor_val["KL"] = []

    lu.print_green("Starting training")

    best_loss = float("inf")

    settings.cyclic_phases

    training_start_time = time()

    for epoch in tqdm(range(settings.cyclic_phases[-1]), desc="Training", ncols=100):

        desc = f"Epoch: {epoch} -- {loss_str}"

        num_elem = len(list_data_train)
        num_batches = num_elem // min(num_elem // 2, settings.batch_size)
        list_batches = np.array_split(np.arange(num_elem), num_batches)
        np.random.shuffle(list_batches)
        for batch_idxs in tqdm(
            list_batches,
            desc=desc,
            ncols=100,
            bar_format="{desc} |{bar}| {n_fmt}/{total_fmt} {rate_fmt}{postfix}",
        ):

            # Sample a batch in packed sequence form
            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                list_data_train, batch_idxs, settings
            )
            # Train step : forward backward pass
            tu.train_step(
                settings,
                rnn,
                packed,
                target_tensor,
                criterion,
                optimizer,
                target_tensor.size(0),
                len(list_batches),
            )

        for param_group in optimizer.param_groups:

            param_group["lr"] = one_cycle_sched(
                epoch, min_learning_rate, max_learning_rate, settings.cyclic_phases
            )

        if (epoch + 1) % settings.monitor_interval == 0:

            # Get metrics (subsample training set to same size as validation set for speed)
            d_losses_train = tu.get_evaluation_metrics(
                settings, list_data_train, rnn, sample_size=len(list_data_val)
            )
            d_losses_val = tu.get_evaluation_metrics(
                settings, list_data_val, rnn, sample_size=None
            )

            # Add current loss avg to list of losses
            for key in d_losses_train.keys():
                d_monitor_train[key].append(d_losses_train[key])
                d_monitor_val[key].append(d_losses_val[key])
            d_monitor_train["epoch"].append(epoch + 1)
            d_monitor_val["epoch"].append(epoch + 1)

            # Prepare loss_str to update progress bar
            loss_str = tu.get_loss_string(d_losses_train, d_losses_val)

            tu.plot_loss(d_monitor_train, d_monitor_val, epoch, settings)
            if d_monitor_val["loss"][-1] < best_loss:
                best_loss = d_monitor_val["loss"][-1]
                torch.save(
                    rnn.state_dict(),
                    f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt",
                )

    training_time = time() - training_start_time

    lu.print_green("Finished training")

    tu.save_training_results(settings, d_monitor_val, training_time)


def save_normalizations(settings):
    """Save normalization used for training

    Saves a json file with the normalization used for each feature

    Arguments:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    dic_norm = {}
    for i, f in enumerate(settings.training_features_to_normalize):
        dic_norm[f] = {}
        for j, w in enumerate(["min", "mean", "std"]):
            dic_norm[f][w] = float(settings.arr_norm[i, j])

    fname = f"{Path(settings.rnn_dir)}/data_norm.json"
    with open(fname, "w") as f:
        json.dump(dic_norm, f, indent=4, sort_keys=True)


def train(settings):
    """Train RNN models

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """
    if settings.swag:
        train_swag(settings)
    elif settings.swag_pretrained_path:
        train_swag_finetune(settings)
    else:
        train_(settings)


def train_(settings):
    """Train RNN models with a decay on plateau policy

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    # save training data config
    save_normalizations(settings)

    # Data
    list_data_train, list_data_val = tu.load_HDF5(settings, test=False)

    # Model specification
    rnn = tu.get_model(settings, len(settings.training_features))
    criterion = nn.CrossEntropyLoss()
    optimizer = tu.get_optimizer(settings, rnn)

    # Prepare for GPU if required
    if settings.use_cuda:
        rnn.cuda()
        criterion.cuda()

    # Keep track of losses for plotting
    loss_str = ""
    d_monitor_train = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    d_monitor_val = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    if "bayesian" in settings.pytorch_model_name:
        d_monitor_train["KL"] = []
        d_monitor_val["KL"] = []

    lu.print_green("Starting training")

    plateau_accuracy = tu.StopOnPlateau(reduce_lr_on_plateau=True)

    best_loss = float("inf")

    training_start_time = time()

    for epoch in tqdm(range(settings.nb_epoch), desc="Training", ncols=100):

        desc = f"Epoch: {epoch} -- {loss_str}"

        num_elem = len(list_data_train)
        num_batches = num_elem // min(num_elem // 2, settings.batch_size)
        list_batches = np.array_split(np.arange(num_elem), num_batches)
        np.random.shuffle(list_batches)
        for batch_idxs in tqdm(
            list_batches,
            desc=desc,
            ncols=100,
            bar_format="{desc} |{bar}| {n_fmt}/{total_fmt} {rate_fmt}{postfix}",
        ):

            # Sample a batch in packed sequence form
            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                list_data_train, batch_idxs, settings
            )
            # Exception for multiclass
            if settings.nb_classes <= int(target_tensor.max()):
                print("")
                lu.print_red(
                    "All sntypes where not defined during database creation (multiclass fails)"
                )
                raise ValueError
            # Train step : forward backward pass
            tu.train_step(
                settings,
                rnn,
                packed,
                target_tensor,
                criterion,
                optimizer,
                target_tensor.size(0),
                len(list_batches),
            )

        if (epoch + 1) % settings.monitor_interval == 0:

            # Get metrics (subsample training set to same size as validation set for speed)
            d_losses_train = tu.get_evaluation_metrics(
                settings, list_data_train, rnn, sample_size=len(list_data_val)
            )
            d_losses_val = tu.get_evaluation_metrics(
                settings, list_data_val, rnn, sample_size=None
            )

            end_condition = plateau_accuracy.step(d_losses_val["Acc"], optimizer)
            if end_condition is True:
                break

            # Add current loss avg to list of losses
            for key in d_losses_train.keys():
                d_monitor_train[key].append(d_losses_train[key])
                d_monitor_val[key].append(d_losses_val[key])
            d_monitor_train["epoch"].append(epoch + 1)
            d_monitor_val["epoch"].append(epoch + 1)

            # Prepare loss_str to update progress bar
            loss_str = tu.get_loss_string(d_losses_train, d_losses_val)

            tu.plot_loss(d_monitor_train, d_monitor_val, epoch, settings)
            if d_monitor_val["loss"][-1] < best_loss:
                best_loss = d_monitor_val["loss"][-1]
                torch.save(
                    rnn.state_dict(),
                    f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt",
                )

    lu.print_green("Finished training")

    training_time = time() - training_start_time

    tu.save_training_results(settings, d_monitor_val, training_time)


def train_swag(settings):
    """Train RNN models with SWAG

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    # save training data config
    save_normalizations(settings)

    # Data
    list_data_train, list_data_val = tu.load_HDF5(settings, test=False)

    # Model specification
    rnn = tu.get_model(settings, len(settings.training_features))
    criterion = nn.CrossEntropyLoss()
    optimizer = tu.get_optimizer(settings, rnn)

    # Initiate SWAG model
    swag_rnn = SwagModel(rnn)
    swag_start = 1

    # Prepare for GPU if required
    if settings.use_cuda:
        rnn.cuda()
        swag_rnn.cuda()
        criterion.cuda()

    # Keep track of losses for plotting
    loss_str = ""
    d_monitor_train = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    d_monitor_val = {"loss": [], "AUC": [], "Acc": [], "epoch": []}

    if "bayesian" in settings.pytorch_model_name:
        d_monitor_train["KL"] = []
        d_monitor_val["KL"] = []

    lu.print_green("Starting training")

    best_loss = float("inf")

    training_start_time = time()

    for epoch in tqdm(range(settings.nb_epoch), desc="Training", ncols=100):

        desc = f"Epoch: {epoch} -- {loss_str}"

        num_elem = len(list_data_train)
        num_batches = num_elem // min(num_elem // 2, settings.batch_size)
        list_batches = np.array_split(np.arange(num_elem), num_batches)
        np.random.shuffle(list_batches)
        for batch_idxs in tqdm(
            list_batches,
            desc=desc,
            ncols=100,
            bar_format="{desc} |{bar}| {n_fmt}/{total_fmt} {rate_fmt}{postfix}",
        ):

            # Sample a batch in packed sequence form
            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                list_data_train, batch_idxs, settings
            )
            # Exception for multiclass
            if settings.nb_classes <= int(target_tensor.max()):
                print("")
                lu.print_red(
                    "All sntypes where not defined during database creation (multiclass fails)"
                )
                raise ValueError
            # Train step : forward backward pass
            tu.train_step(
                settings,
                rnn,
                packed,
                target_tensor,
                criterion,
                optimizer,
                target_tensor.size(0),
                len(list_batches),
            )

        if (epoch + 1) % settings.monitor_interval == 0:

            # Get metrics (subsample training set to same size as validation set for speed)
            d_losses_train = tu.get_evaluation_metrics(
                settings, list_data_train, rnn, sample_size=len(list_data_val)
            )
            d_losses_val = tu.get_evaluation_metrics(
                settings, list_data_val, rnn, sample_size=None
            )

            # Add current loss avg to list of losses
            for key in d_losses_train.keys():
                d_monitor_train[key].append(d_losses_train[key])
                d_monitor_val[key].append(d_losses_val[key])
            d_monitor_train["epoch"].append(epoch + 1)
            d_monitor_val["epoch"].append(epoch + 1)

            # Prepare loss_str to update progress bar
            loss_str = tu.get_loss_string(d_losses_train, d_losses_val)

            tu.plot_loss(d_monitor_train, d_monitor_val, epoch, settings)
            if d_monitor_val["loss"][-1] < best_loss:
                best_loss = d_monitor_val["loss"][-1]
                torch.save(
                    rnn.state_dict(),
                    f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt",
                )

        if epoch >= settings.swag_start_epoch:
            if swag_start == 1:
                lu.print_green(
                    "Starting SWAG process: calculating SWA and SWAG-diagonal"
                )
                swag_start = 0

            swag_rnn.update_parameters(rnn)

            # start SWAG samplling when K > 1, i.e. swag_rnn.n_averaged > 2
            if epoch >= settings.swag_start_epoch + 2:
                lu.print_green("Starting SWAG sampling")
                d_losses_train_swag = tu.get_evaluation_metrics(
                    settings,
                    list_data_train,
                    swag_rnn,
                    sample_size=len(list_data_val),
                    swag_sampling=True,
                )
                print("swag metrics:", d_losses_train_swag)
                torch.save(
                    swag_rnn,
                    f"{settings.rnn_dir}/{settings.pytorch_model_name}_swag.pt",
                )

    lu.print_green("Finished training")

    training_time = time() - training_start_time

    tu.save_training_results(settings, d_monitor_val, training_time)


def train_swag_finetune(settings):
    """Train RNN models with SWAG

    Args:
        settings (ExperimentSettings): controls experiment hyperparameters
    """

    # save training data config
    save_normalizations(settings)

    # Data
    list_data_train, list_data_val = tu.load_HDF5(settings, test=False)

    # Model specification
    rnn = tu.get_model(settings, len(settings.training_features))
    
    # Load pretrained model
    lu.print_green(
        f"Loading pretrained SWAG model from {settings.swag_pretrained_path}"
    )
    rnn.load_state_dict(
        torch.load(settings.swag_pretrained_path, map_location="cpu", weights_only=False)
    )
    
    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        rnn.parameters(),
        lr=settings.learning_rate,
        momentum=0.9, # TODO do not hardcode
        weight_decay=1e-4, # TODO do not hardcode
    )

    settings.swag_start_epoch = 0 # TODO do not hardcode
    settings.nb_epoch = 30 # TODO do not hardcode

    # Initiate SWAG model
    swag_rnn = SwagModel(rnn)

    # Prepare for GPU if required
    if settings.use_cuda:
        rnn.cuda()
        swag_rnn.cuda()
        criterion.cuda()

    # Keep track of losses for plotting
    loss_str = ""
    d_monitor_train = {"loss": [], "AUC": [], "Acc": [], "epoch": []}
    d_monitor_val = {"loss": [], "AUC": [], "Acc": [], "epoch": []}

    if "bayesian" in settings.pytorch_model_name:
        d_monitor_train["KL"] = []
        d_monitor_val["KL"] = []

    lu.print_green("Starting training")

    best_loss = float("inf")

    training_start_time = time()

    for epoch in tqdm(range(settings.nb_epoch), desc="Training", ncols=100):

        desc = f"Epoch: {epoch} -- {loss_str}"

        num_elem = len(list_data_train)
        num_batches = num_elem // min(num_elem // 2, settings.batch_size)
        list_batches = np.array_split(np.arange(num_elem), num_batches)
        np.random.shuffle(list_batches)
        for batch_idxs in tqdm(
            list_batches,
            desc=desc,
            ncols=100,
            bar_format="{desc} |{bar}| {n_fmt}/{total_fmt} {rate_fmt}{postfix}",
        ):

            # Sample a batch in packed sequence form
            packed, _, target_tensor, idxs_rev_sort = tu.get_data_batch(
                list_data_train, batch_idxs, settings
            )
            # Exception for multiclass
            if settings.nb_classes <= int(target_tensor.max()):
                print("")
                lu.print_red(
                    "All sntypes where not defined during database creation (multiclass fails)"
                )
                raise ValueError
            # Train step : forward backward pass
            tu.train_step(
                settings,
                rnn,
                packed,
                target_tensor,
                criterion,
                optimizer,
                target_tensor.size(0),
                len(list_batches),
            )

        if (epoch + 1) % settings.monitor_interval == 0:

            # Get metrics (subsample training set to same size as validation set for speed)
            d_losses_train = tu.get_evaluation_metrics(
                settings, list_data_train, rnn, sample_size=len(list_data_val)
            )
            d_losses_val = tu.get_evaluation_metrics(
                settings, list_data_val, rnn, sample_size=None
            )

            # Add current loss avg to list of losses
            for key in d_losses_train.keys():
                d_monitor_train[key].append(d_losses_train[key])
                d_monitor_val[key].append(d_losses_val[key])
            d_monitor_train["epoch"].append(epoch + 1)
            d_monitor_val["epoch"].append(epoch + 1)

            # Prepare loss_str to update progress bar
            loss_str = tu.get_loss_string(d_losses_train, d_losses_val)

            tu.plot_loss(d_monitor_train, d_monitor_val, epoch, settings)
            if d_monitor_val["loss"][-1] < best_loss:
                best_loss = d_monitor_val["loss"][-1]
                torch.save(
                    rnn.state_dict(),
                    f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt",
                )

        lu.print_green("Updating SWAG estimators")
        swag_rnn.update_parameters(rnn)

        if epoch >= 2:
            lu.print_green("Starting SWAG sampling")
            d_losses_train_swag = tu.get_evaluation_metrics(
                settings,
                list_data_train,
                swag_rnn,
                sample_size=len(list_data_val),
                swag_sampling=True,
            )
            print("swag metrics:", d_losses_train_swag)
            torch.save(
                swag_rnn,
                f"{settings.rnn_dir}/{settings.pytorch_model_name}_swag.pt",
            )

    lu.print_green("Finished training")

    training_time = time() - training_start_time

    tu.save_training_results(settings, d_monitor_val, training_time)



import torch


class VariationalRNN(torch.nn.Module):
    def __init__(self, input_size, settings):

        super(VariationalRNN, self).__init__()

        # Params
        self.input_size = input_size
        self.layer_type = settings.layer_type
        self.output_size = settings.nb_classes
        self.hidden_size = settings.hidden_dim
        self.num_layers = settings.num_layers
        self.bidirectional = settings.bidirectional
        self.dropout = settings.dropout
        self.use_cuda = settings.use_cuda
        self.rnn_output_option = settings.rnn_output_option

        bidirectional_factor = 2 if self.bidirectional is True else 1
        last_input_size = (
            self.hidden_size * bidirectional_factor
            if self.rnn_output_option == "mean"
            else self.hidden_size * bidirectional_factor * self.num_layers
        )

        # Need to create recurrent layers one by one because we use variational dropout
        self.rnn_layers = []
        for i in range(self.num_layers):
            if i == 0:
                input_size = self.input_size
            else:
                previous_layer = getattr(self, f"rnn_layer{i - 1}")
                input_size = previous_layer.module.hidden_size * bidirectional_factor

            # Create recurrent layer
            layer = getattr(torch.nn, self.layer_type.upper())(
                input_size,
                self.hidden_size,
                num_layers=1,
                dropout=0.0,  # no dropout: we later create a specific layer for that
                bidirectional=self.bidirectional,
            )
            # Apply weight drop
            layer = WeightDrop(layer, ["weight_hh_l0"], dropout=self.dropout)
            # Set layer as attribute
            setattr(self, f"rnn_layer{i}", layer)
            self.rnn_layers.append(layer)

        self.recurrent_dropout_layer = VariationalRecurrentDropout()
        self.output_dropout_layer = VariationalDropout()
        self.output_layer = torch.nn.Linear(last_input_size, self.output_size)

    def apply_recurrent_variational_dropout(
        self, x, dropout_value, mean_field_inference=False
    ):

        # apply dropout to input
        if isinstance(x, torch.nn.utils.rnn.PackedSequence):
            # Unpack
            x, lengths = torch.nn.utils.rnn.pad_packed_sequence(x)
            # Apply locked dropout
            x = self.recurrent_dropout_layer(
                x, dropout_value, mean_field_inference=mean_field_inference
            )
            # Repack
            x = torch.nn.utils.rnn.pack_padded_sequence(x, lengths)
        else:
            x = self.recurrent_dropout_layer(
                x, dropout_value, mean_field_inference=mean_field_inference
            )

        return x

    def forward(self, x, mean_field_inference=False):

        # apply variational dropout to input
        x = self.apply_recurrent_variational_dropout(
            x, self.dropout, mean_field_inference
        )

        list_hidden = []
        for layer_idx, rnn_layer in enumerate(self.rnn_layers):
            x, hidden = rnn_layer(x, mean_field_inference=mean_field_inference)
            list_hidden.append(hidden)

            # Apply Variational dropout between recurrent layers
            if layer_idx != len(self.rnn_layers) - 1:
                x = self.apply_recurrent_variational_dropout(
                    x, self.dropout, mean_field_inference=mean_field_inference
                )

        # Output options
        # Standard: all layers, only end of pass
        #    - take last pass in all layers (hidden)
        #    - reshape and apply dropout
        #    - use h20 to obtain output (h2o input: hidden_size*num_layers*bi)
        # Mean: last layer, mean on sequence
        #    - take packed output from last layer (out) that contains all time steps for the last layer
        #    - find where padding was done and create a mask for those values, apply this mask
        #    - take a mean for the whole sequence (time_steps)
        #    - use h2o to obtain output (beware! it is only one layer deep since it is the last one only)

        if self.rnn_output_option == "standard":
            # Special case for lstm where hidden = (h, c)

            # In vanilla standard we have: out, hidden = rnn(X)
            # hidden is built as follows:
            # for each layer, for each direction, take the last hidden state
            # concatenate the results. obtain hidden (num_layers * num_directions, B, D)
            # here, we need to do it manually as the recurrent layers are written one by one
            # We now carry out the ``concatenate`` operation.
            if self.layer_type == "lstm":
                hn = torch.cat([h[0] for h in list_hidden], dim=0)
            else:
                hn = torch.cat(list_hidden, dim=0)

            hn = hn.permute(1, 2, 0).contiguous()
            # hn is (num_layers * num_directions, batch, hidden_size)
            batch_size = hn.shape[0]
            # hn now is (batch, hidden size, num_layers * num_directions)
            x = hn.view(batch_size, -1)
            # x is (batch, hidden size * num_layers * num_directions)

        if self.rnn_output_option == "mean":
            if isinstance(x, torch.nn.utils.rnn.PackedSequence):
                x, lens = torch.nn.utils.rnn.pad_packed_sequence(x)
                # x is (seq_len, batch, hidden size)

                # take mean over seq_len
                x = x.sum(0) / lens.unsqueeze(-1).float().to(x.device)
                # x is (batch, hidden_size)
            else:
                x = x.mean(0)
                # x is (batch, hidden_size)

        # apply dropout
        x = self.output_dropout_layer(
            x, self.dropout, mean_field_inference=mean_field_inference
        )
        # Final projection layer
        output = self.output_layer(x)

        return output


class VariationalRecurrentDropout(torch.nn.Module):
    """
    This is a renamed Locked Dropout from https://github.com/salesforce/awd-lstm-lm

    - We added a mean_field_inference flag to carry out mean field inference

    We do this so that choosing whether to use dropout or not at inference time
    is explicit

    """

    def __init__(self):
        super().__init__()

    def forward(self, x, dropout, mean_field_inference=False):
        if mean_field_inference is True:
            return x
        else:
            m = x.data.new(1, x.size(1), x.size(2)).bernoulli_(1 - dropout)
            mask = m / (1 - dropout)  # rescaling to account for dropout
            mask = mask.expand_as(x)

            return mask * x


class VariationalDropout(torch.nn.Module):
    """Re-implementation of torch.nn.modules.Dropout

    - training flag is always set to True
    - We added a mean_field_inference flag to carry out mean field inference

    We do this rather than using the training flag so that we can more explicitly decide
    to use dropout or not at inference time
    """

    def forward(self, x, dropout, mean_field_inference=False):
        training = True
        inplace = False
        if mean_field_inference is True:
            return x
        else:
            return torch.nn.functional.dropout(x, dropout, training, inplace)


class WeightDrop(torch.nn.Module):
    """
    WeightDrop from https://github.com/salesforce/awd-lstm-lm

    - Removed the variational input parameter as we will always use WeightDrop in variational mode

    """

    def __init__(self, module, weights, dropout):
        super(WeightDrop, self).__init__()
        self.module = module
        self.weights = weights
        self.dropout = dropout
        self._setup()

    def dummy_function(*args, **kwargs):
        # We need to replace flatten_parameters with a nothing function
        # It must be a function rather than a lambda as otherwise pickling explodes
        return

    def _setup(self):
        # Terrible temporary solution to an issue regarding compacting weights re: CUDNN RNN
        if issubclass(type(self.module), torch.nn.RNNBase):
            self.module.flatten_parameters = self.dummy_function

        for name_w in self.weights:
            print(f"Applying weight drop of {self.dropout} to {name_w}")
            w = getattr(self.module, name_w)  # take this param of the module
            del self.module._parameters[name_w]
            self.module.register_parameter(
                name_w + "_raw", torch.nn.Parameter(w.data)
            )  # recreate same data with different name

    def _setweights(self, mean_field_inference=False):
        for name_w in self.weights:
            raw_w = getattr(self.module, name_w + "_raw")
            w = None
            mask = torch.ones(raw_w.size(0), 1)
            device = raw_w.device
            mask = mask.to(device)
            # forcing it to be applied training and validation
            if mean_field_inference is False:
                mask = torch.nn.functional.dropout(mask, p=self.dropout, training=True)
            mask = mask.expand_as(raw_w)
            # applying same mask to all elements of the batch
            # h [batch, xdim]
            w = mask * raw_w
            setattr(self.module, name_w, w)

    def forward(self, *args, mean_field_inference=False):
        # Apply dropout to weights of module
        self._setweights(mean_field_inference=mean_field_inference)

        return self.module.forward(*args)


def embedded_dropout(embed, words, dropout, mean_field_inference=False):
    """
    embedded_dropout from https://github.com/salesforce/awd-lstm-lm
    with some modifications like the mean field inference flag

    """
    if mean_field_inference:
        masked_embed_weight = embed.weight
    else:
        mask = embed.weight.data.new().resize_((embed.weight.size(0), 1)).bernoulli_(
            1 - dropout
        ).expand_as(embed.weight) / (1 - dropout)
        masked_embed_weight = mask * embed.weight

    padding_idx = embed.padding_idx
    if padding_idx is None:
        padding_idx = -1

    X = torch.nn.functional.embedding(
        words,
        masked_embed_weight,
        padding_idx,
        embed.max_norm,
        embed.norm_type,
        embed.scale_grad_by_freq,
        embed.sparse,
    )

    return X



import re
import h5py
import torch
import numpy as np
import pandas as pd
import torch.nn as nn
from tqdm import tqdm
from pathlib import Path
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from ..utils import data_utils as du
from ..utils import logging_utils as lu
from ..utils import training_utils as tu
from ..utils.visualization_utils import FILTER_COLORS, ALL_COLORS, LINE_STYLE

plt.switch_backend("agg")


def get_predictions(settings, dict_rnn, X, target, OOD=None):

    list_data = [(X.copy(), target)]

    _, X_tensor, *_ = tu.get_data_batch(list_data, [0], settings, OOD=OOD)

    if settings.use_cuda:
        X_tensor.cuda()

    d_pred = {key: {"prob": []} for key in dict_rnn}

    # Apply rnn to obtain prediction
    for model_type, rnn in dict_rnn.items():

        n = settings.num_inference_samples if "variational" in model_type else 1
        new_size = (X_tensor.size(0), n, X_tensor.size(2))

        if "bayesian" in model_type:

            # Loop over num samples to obtain predictions
            list_out = [
                rnn(X_tensor.expand(new_size))
                for i in range(settings.num_inference_samples)
            ]
            out = torch.cat(list_out, dim=0)
            # Apply softmax to obtain a proba
            pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
        else:
            out = rnn(X_tensor.expand(new_size))
            # Apply softmax to obtain a proba
            pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

        # Add to buffer list
        d_pred[model_type]["prob"].append(pred_proba)

    # Stack
    for key in dict_rnn.keys():
        arr_proba = np.stack(d_pred[key]["prob"], axis=0)
        d_pred[key]["prob"] = arr_proba  # arr_prob is (T, num_samples, 2)

    return d_pred, X_tensor.squeeze().detach().cpu().numpy()


def plot_distributions(settings, list_d_plot):

    plt.figure(figsize=(20, 30))
    gs = gridspec.GridSpec(8, 2, hspace=0.3, wspace=0.1)

    for i in range(len(list_d_plot)):

        d_plot = list_d_plot[i]
        SNID = d_plot["SNID"]
        OOD = d_plot["OOD"]
        target = d_plot["target"]
        redshift = d_plot["redshift"]
        peak_MJD = d_plot["peak_MJD"]
        d_pred = d_plot["d_pred"]

        # Plot the lightcurve
        ax = plt.subplot(gs[2 * i])
        for flt in settings.list_filters:
            flt_time = d_plot[flt]["MJD"]
            # Only plot a time series if it's non empty
            if len(flt_time) > 0:
                flux = d_plot[flt]["FLUXCAL"]
                fluxerr = d_plot[flt]["FLUXCALERR"]
                ax.errorbar(
                    flt_time,
                    flux,
                    yerr=fluxerr,
                    fmt="o",
                    label=f"Filter {flt}",
                    color=FILTER_COLORS[flt],
                )

        ax.set_ylabel("FLUXCAL", fontsize=24)
        if i == 7:
            ax.set_xlabel("days", fontsize=24)
        ylim = ax.get_ylim()
        plt.yticks(fontsize=14)
        plt.xticks(fontsize=14)

        SNtype = du.sntype_decoded(target, settings)
        if OOD is not None:
            ax.set_title(f"OOD {OOD} ID: {SNID}", fontsize=24)
        else:
            ax.set_title(
                SNtype + f" (ID: {SNID}, redshift: {redshift:.3g})", fontsize=24
            )
            # Add PEAKMJD
            ax.plot([peak_MJD, peak_MJD], ylim, "k--", label="Peak MJD")

        # Plot the classifications
        ax = plt.subplot(gs[2 * i + 1])
        ax.set_xlim(-0.1, 1.1)

        for idx, key in enumerate(d_pred.keys()):

            arr_prob = np.squeeze(d_pred[key]["prob"])
            all_probs = np.ravel(arr_prob)
            _, bin_edges = np.histogram(all_probs, bins=25)

            for class_prob in range(settings.nb_classes):
                color = ALL_COLORS[class_prob + idx * settings.nb_classes]
                label = du.sntype_decoded(class_prob, settings)
                linestyle = LINE_STYLE[class_prob]

                if len(d_pred) > 1:
                    label += f" {key}"

                ax.hist(
                    arr_prob[:, class_prob],
                    color=color,
                    histtype="step",
                    linestyle=linestyle,
                    linewidth=2,
                    label=label,
                    bins=bin_edges,
                )
        plt.yticks(fontsize=14)
        plt.xticks(fontsize=14)

        ax.set_yscale("log")

        if i == len(list_d_plot) - 1:
            ax.legend(
                bbox_to_anchor=(-2.8, 5.8),
                loc=2,
                borderaxespad=0.0,
                fontsize=16,
                ncol=(settings.nb_classes // 2) + 1,
                framealpha=0,
            )
        if 2 * i + 1 == 15:
            ax.set_xlabel("classification probability", fontsize=24)

    plt.subplots_adjust(
        left=0.08, right=0.99, bottom=0.03, top=0.98, wspace=0.0, hspace=0.02
    )

    if len([settings.model_files]) == 1:
        parent_dir = Path(settings.model_files[0]).parent.name
        fig_path = f"{settings.lightcurves_dir}/{parent_dir}/prediction_distribution"
        fig_name = f"{parent_dir}.png"
    else:
        fig_path = f"{settings.lightcurves_dir}/{settings.pytorch_model_name}/prediction_distribution"
        fig_name = f"{settings.pytorch_model_name}.png"
    Path(fig_path).mkdir(parents=True, exist_ok=True)
    # plt.tight_layout()
    plt.savefig(Path(fig_path) / fig_name)
    plt.clf()
    plt.close()


def plot_prediction_distribution(settings):
    """Load model corresponding to settings
    or (if specified) load a list of models.

    Args:
        settings: (ExperimentSettings) custom class to hold hyperparameters
        int (nb_lcs): number of light-curves to plot, default is 1
    """

    # No plot for ternary tasks
    if settings.nb_classes == 3:
        return

    settings.random_length = False
    settings.random_redshift = False

    # Load the test data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Load features list
    file_name = f"{settings.processed_dir}/database.h5"
    with h5py.File(file_name, "r") as hf:
        features = hf["features"][settings.idx_features]
        # Convert byte strings to Unicode strings if necessary
        if isinstance(features[0], bytes):
            features = [feature.decode("utf-8") for feature in features]

    # Load RNN model
    dict_rnn = {}
    if settings.model_files is None:
        settings.model_files = [f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt"]
    else:
        assert (
            len(settings.model_files) == 1
        ), "Only one model file allowed at a time for these plots"

    # Check that the settings match the model file
    base_files = [Path(f).name for f in settings.model_files]

    classes = [int(re.search(r"(?<=CLF\_)\d+(?=\_)", f).group()) for f in base_files]
    redshifts = [re.search(r"(?<=R\_)[A-Za-z]+(?=\_)", f).group() for f in base_files]

    nb_classes, redshift = classes[0], redshifts[0]
    assert settings.nb_classes == nb_classes, lu.str_to_redstr(
        "Incompatible nb_classes between CLI and model files"
    )
    assert str(settings.redshift) == redshift, lu.str_to_redstr(
        "Incompatible redshift between CLI and model files"
    )

    for model_file in settings.model_files:
        if "variational" in model_file:
            settings.model = "variational"
        if "vanilla" in model_file:
            settings.model = "vanilla"
        if "bayesian" in model_file:
            settings.model = "bayesian"
        rnn = tu.get_model(settings, len(settings.training_features))
        rnn_state = torch.load(
            model_file,
            map_location=lambda storage, loc: storage,
            weights_only=False,
        )
        rnn.load_state_dict(rnn_state)
        rnn.to(settings.device)
        rnn.eval()
        name = (
            f"{settings.model} photometry"
            if "photometry" in model_file
            else f"{settings.model} salt"
        )
        dict_rnn[name] = rnn

    # lOad SN info
    SNinfo_df = du.load_HDF5_SNinfo(settings)

    targets = np.array([o[1] for o in list_data_test])

    if settings.nb_classes == 2:
        # 2 lightcurves of each type
        idxs_keep = (
            np.where(targets == 0)[0][:2].tolist()
            + np.where(targets == 1)[0][:2].tolist()
        )
    elif settings.nb_classes == 7:
        # Ia, Ib, Ic, IIp
        idxs_keep = [
            np.where(targets == list(settings.sntypes.values()).index(i))[0][0]
            for i in ["Ia", "Ib", "Ic", "IIP"]
        ]

    # Carry out 8 plots: 4 real light curves + 4 OOD
    list_OOD_types = ["random", "sin", "reverse", "shuffle"]
    list_data_test = [(list_data_test[i], None) for i in idxs_keep]
    list_data_test += [(o[0], ood) for (o, ood) in zip(list_data_test, list_OOD_types)]

    list_d_plot = []

    # Loop over data to plot prediction
    for ((X, target, SNID, _, X_ori), OOD) in tqdm(list_data_test, ncols=100):

        redshift = SNinfo_df[SNinfo_df["SNID"] == SNID]["SIM_REDSHIFT_CMB"].values[0]
        peak_MJD = SNinfo_df[SNinfo_df["SNID"] == SNID]["PEAKMJDNORM"].values[0]

        # Prepare plotting data in a dict
        d_plot = {
            flt: {"FLUXCAL": [], "FLUXCALERR": [], "MJD": []}
            for flt in settings.list_filters
        }

        with torch.no_grad():
            d_pred, X_normed = get_predictions(settings, dict_rnn, X, target, OOD=OOD)
        # X here has been normalized. We unnormalize X
        X_unnormed = tu.unnormalize_arr(X_normed, settings)
        # Check we do recover X_ori when OOD is None
        if OOD is None:
            assert np.all(np.isclose(np.ravel(X_ori), np.ravel(X_unnormed), atol=1e-2))

        # TODO: IMPROVE
        df_temp = pd.DataFrame(data=X_unnormed, columns=features)
        arr_time = np.cumsum(df_temp.delta_time.values)
        for flt in settings.list_filters:
            non_zero = np.where(
                ~np.isclose(df_temp[f"FLUXCAL_{flt}"].values, 0, atol=1e-2)
            )[0]
            d_plot[flt]["FLUXCAL"] = df_temp[f"FLUXCAL_{flt}"].values[non_zero]
            d_plot[flt]["FLUXCALERR"] = df_temp[f"FLUXCALERR_{flt}"].values[non_zero]
            d_plot[flt]["MJD"] = arr_time[non_zero]

        d_plot["redshift"] = redshift
        d_plot["peak_MJD"] = peak_MJD
        d_plot["SNID"] = SNID
        d_plot["OOD"] = OOD
        d_plot["target"] = target
        d_plot["d_pred"] = d_pred

        list_d_plot.append(d_plot)

    plot_distributions(settings, list_d_plot)
    lu.print_green("Finished plotting lightcurves and predictions ")



import os
import re
import numpy as np
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt

from ..utils import data_utils as du
from ..utils import performance_utils as pu
from ..utils import logging_utils as lu

plt.switch_backend("agg")


def aggregate_metrics(settings):
    """Aggregate all pre-computed METRICS files into a single dataframe
    for analysis

    Save a csv dataframe aggregating all the metrics

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters

    """

    list_files = Path(f"{settings.models_dir}").glob("**/*METRICS*.pickle")
    list_files = list(map(str, list_files))
    assert len(list_files) != 0, lu.str_to_redstr(
        "No predictions found. Please train and validate vanilla models"
    )

    # read all performance metrics
    list_df = []
    for f in list_files:
        df = pd.read_pickle(f)
        model_name = df["model_name"][0]
        source_data = df["source_data"][0]
        model_name_noseed = re.sub(r"S\_\d+_", "", model_name)
        model_name_noseed = f"{model_name_noseed}"
        df["model_name_noseed"] = model_name_noseed
        df["source_data"] = source_data
        list_df.append(df)

    df_all = pd.concat(list_df, axis=0, sort=True)

    # Groupby model and average over seed
    group_cols = ["model_name_noseed", "source_data"]
    df_mean = df_all.groupby(group_cols).mean().add_suffix("_mean")
    df_std = df_all.groupby(group_cols).std().add_suffix("_std")

    mean_columns = df_mean.columns
    std_columns = df_std.columns

    orderded_columns = [
        item for sublist in zip(mean_columns, std_columns) for item in sublist
    ]
    orderded_columns = ["model_name_noseed", "source_data"] + orderded_columns

    df_stats = pd.concat([df_mean, df_std], axis=1).reset_index()[orderded_columns]
    df_stats.to_csv(Path(settings.stats_dir) / "summary_stats.csv", index=False)


def get_metrics_singlemodel(settings, prediction_file=None, model_type="rnn"):
    """Launch computation of all evaluation metrics for a given model, specified
    by the settings object or by a model file

    Save a pickled dataframe (we pickle  because we're saving numpy arrays, which
    are not easily savable with the ``to_csv`` method).

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        prediction_file (str): Path to saved predictions. Default: ``None``
        model_type (str): only ``rnn`` available for the moment

    Returns:
        (pandas.DataFrame) holds the performance metrics for this dataframe
    """
    # list current support model type
    supported_models = ["rnn"]

    # check whether input model type is supported
    if model_type not in supported_models:
        raise ValueError(
            f"Unsupported model type '{model_type}'. Currently supported models: {', '.join(supported_models)}"
        )

    df_SNinfo = du.load_HDF5_SNinfo(settings)
    if Path(f"{settings.processed_dir}/hostspe_SNID.csv").exists():
        host = pd.read_csv(f"{settings.processed_dir}/hostspe_SNID.csv")
    else:
        host = pd.read_pickle(f"{settings.processed_dir}/hostspe_SNID.pickle")
    host_zspe_list = host["SNID"].tolist()

    if prediction_file is not None:
        # Overwrite dump_dir: use the one corresponding to specified model_file
        # Useful for representativeness studies
        dump_dir = str(Path(prediction_file).parent)
        # Also overwrite model name
        model_name = Path(prediction_file).parent.name
        metrics_file = str(prediction_file).replace("PRED_", "METRICS_")
        source_data = (
            "photometry" if "photometry" in Path(prediction_file).name else "saltfit"
        )
    else:
        if model_type == "rnn":
            model_name = settings.pytorch_model_name
        dump_dir = f"{settings.models_dir}/{model_name}"
        prediction_file = f"{dump_dir}/" f"PRED_{model_name}.pickle"
        metrics_file = f"{dump_dir}/" f"METRICS_{model_name}.pickle"
        source_data = settings.source_data

    assert os.path.isfile(prediction_file), lu.str_to_redstr(
        f"{prediction_file} DOES NOT EXIST"
    )

    df = pd.read_pickle(prediction_file)
    df = pd.merge(df, df_SNinfo[["SNID", settings.sntype_var]], on="SNID", how="left")

    list_df_metrics = []

    # Metrics of RNN
    list_df_metrics.append(get_calibration_metrics_singlemodel(df))

    if model_type == "rnn":
        # RNN-specific metrics
        list_df_metrics.append(
            get_rnn_performance_metrics_singlemodel(settings, df, host_zspe_list)
        )
        list_df_metrics.append(get_uncertainty_metrics_singlemodel(df))
        list_df_metrics.append(get_entropy_metrics_singlemodel(df, settings.nb_classes))
        list_df_metrics.append(
            get_classification_stats_singlemodel(df, settings.nb_classes)
        )

    df_metrics = pd.concat(list_df_metrics, axis=1)

    df_metrics["model_name"] = model_name
    df_metrics["source_data"] = source_data
    df_metrics.to_pickle(metrics_file, protocol=4)

    lu.print_green("Finished getting metrics ")


def get_rnn_performance_metrics_singlemodel(settings, df, host_zspe_list):
    """Compute performance metrics (accuracy, AUC, purity etc) for
    an RNN model

    - Compute metrics around peak light (i.e. ``PEAKMJD``) and for the full lightcurve.
    - For bayesian models, compute multiple predictions per lightcurve and then take the median

    Args:
        settings (ExperimentSettings): custom class to hold hyperparameters
        df (pandas.DataFrame): dataframe containing a model's predictions
        host_zspe_list (list): available host galaxy spectroscopic redshifts

    Returns:
        (pandas.DataFrame) holds the performance metrics for this dataframe
    """

    # Compute metrics around peak light, and with full lightcurve
    list_keys = ["-7", "-2", "", "+2", "+30"] + ["all"]
    perf_dic = {}
    for key in list_keys:
        # Need to select data (remove NAN) as sometimes, MJD happens too early
        # and MJD+(-2) (for instance) cannot be computed, hence NaN
        if key != "all":
            format_key = f"PEAKMJD{key}"
        else:
            format_key = key
        selection = df[~np.isnan(df[f"{format_key}_class1"])]
        if "bayesian" or "variational" in settings.pytorch_model_name:
            group_bayesian = True
        else:
            group_bayesian = False
        # general metrics
        # TODO refactor
        reformatted_selection = pu.reformat_df(
            selection, key, settings, group_bayesian=group_bayesian
        )
        accuracy, auc, purity, efficiency, _ = pu.performance_metrics(
            reformatted_selection
        )
        contamination_df = pu.contamination_by_SNTYPE(reformatted_selection, settings)

        if key == "":
            savekey = "0"
        else:
            savekey = key
        perf_dic[f"{savekey}_accuracy"] = accuracy
        perf_dic[f"{savekey}_auc"] = auc
        perf_dic[f"{savekey}_purity"] = purity
        perf_dic[f"{savekey}_efficiency"] = efficiency
        for sntype, contamination_percentage in contamination_df.values:
            perf_dic[
                f"{savekey}_contamination_{int(sntype)}"
            ] = contamination_percentage

        # Reweighted for SNe with zspe
        zspe_df = selection[selection["SNID"].isin(host_zspe_list)]
        if len(zspe_df) > 0:
            zspe_df = pu.reformat_df(
                zspe_df, key, settings, group_bayesian=group_bayesian
            )
            (
                accuracy_zspe,
                auc_zspe,
                purity_zspe,
                efficiency_zspe,
                _,
            ) = pu.performance_metrics(zspe_df)
        else:
            accuracy_zspe, auc_zspe, purity_zspe, efficiency_zspe = (0.0, 0.0, 0.0, 0.0)

        perf_dic[f"{savekey}_zspe_accuracy"] = accuracy_zspe
        perf_dic[f"{savekey}_zspe_auc"] = auc_zspe
        perf_dic[f"{savekey}_zspe_purity"] = purity_zspe
        perf_dic[f"{savekey}_zspe_efficiency"] = efficiency_zspe

    # Create a dataframe where the columns are the keys of perf_dic
    df_perf = pd.DataFrame.from_dict(perf_dic, orient="index").transpose()

    return df_perf


def get_uncertainty_metrics_singlemodel(df):
    """For any lightcurve, compute the standard deviation of the model's
    predictions (this is only valid for bayesian models which yield
    a distribution of predictions).

    Then, compute the mean and std dev of this distribution across all lightcurves
    A higher mean indicates a model which is less confident in its predictions

    Args:
        df (pandas.DataFrame): dataframe containing a model's predictions

    Returns:
        (pandas.DataFrame) holds the uncertainty metrics for this dataframe
    """

    columns = ["SNID", "all_class0"] + [f"all_{OOD}_class0" for OOD in du.OOD_TYPES]

    g = df[columns].groupby("SNID").std()

    mean_std_dev = g.mean()
    std_std_dev = g.std()

    df_mean = (
        pd.DataFrame(
            data=mean_std_dev.values.reshape(1, -1), columns=mean_std_dev.index.values
        )
        .add_prefix("mean_")
        .add_suffix("_std_dev")
    )

    df_std = (
        pd.DataFrame(
            data=std_std_dev.values.reshape(1, -1), columns=std_std_dev.index.values
        )
        .add_prefix("std_")
        .add_suffix("_std_dev")
    )

    df_uncertainty = pd.concat([df_mean, df_std], axis=1)

    return df_uncertainty


def get_entropy_metrics_singlemodel(df, nb_classes):
    """Compute the entropy of the predictions
    Low entropy indicates a model that is very confident of its predictions

    Args:
        df (pandas.DataFrame): dataframe containing a model's predictions
        nb_classes (int): the number of classes in the classification task

    Returns:
        (pandas.DataFrame) holds the entropy metrics for this dataframe
    """

    list_prefixes = ["all"] + [f"all_{OOD}" for OOD in du.OOD_TYPES]
    list_data = []

    for prefix in list_prefixes:
        list_columns = [f"{prefix}_class{i}" for i in range(nb_classes)]

        arr_proba = df[list_columns].values
        entropy = -(np.log(arr_proba) * arr_proba).sum(axis=-1).mean()

        data = np.array([entropy]).reshape(1, -1)
        list_data.append(data)

    data = np.concatenate(list_data, axis=-1)
    df_entropy = pd.DataFrame(
        data, columns=[f"{prefix}_entropy" for prefix in list_prefixes]
    )

    return df_entropy


def get_calibration_metrics_singlemodel(df):
    """Compute probability calibration dataframe.
    If the calibration curve is close to identity, the model is considered
    well-calibrated.

    Args:
        df (pandas.DataFrame): dataframe containing a model's predictions

    Returns:
        (pandas.DataFrame) holds the calibration metrics for this dataframe
    """

    # TODO: clarify
    bins = np.arange(0, 11) / 10
    df["calibration_TPF"] = df["target"] != (df["all_class0"] < 0.5)
    df["prob_bin"] = pd.cut(df.all_class0.values, bins=bins, labels=range(10))
    df_calib = df[["calibration_TPF", "prob_bin"]].groupby("prob_bin").mean()

    df_calib.loc[df_calib.index >= 5] = 1 - df_calib.loc[df_calib.index >= 5]

    df_calib = df_calib.reset_index()

    # Add mean bins
    df_calib["calibration_mean_bins"] = (
        df[["all_class0", "prob_bin"]].groupby("prob_bin").mean()["all_class0"]
    )

    # Add dispersion
    dispersion = (
        (df_calib["calibration_mean_bins"] - df_calib["calibration_TPF"]) ** 2
    ).mean()

    df_calib_flat = pd.DataFrame([dispersion], columns=["calibration_dispersion"])
    for col in ["calibration_mean_bins", "calibration_TPF"]:
        df_calib_flat[col] = [df_calib[col].values]

    return df_calib_flat


def get_classification_stats_singlemodel(df, nb_classes):
    """Find out how many lightcurves are classified in each class

    Args:
        df (pandas.DataFrame): dataframe containing a model's predictions
        nb_classes (int): the number of classes in the classification task

    Returns:
        (pandas.DataFrame) holds the calibration metrics for this dataframe
    """

    list_prefixes = ["all"] + [f"all_{OOD}" for OOD in du.OOD_TYPES]
    list_df = []

    for prefix in list_prefixes:
        list_columns = [f"{prefix}_class{i}" for i in range(nb_classes)]

        arr_preds = df[list_columns].values
        pred_class = np.argmax(arr_preds, axis=1)
        list_clf_stats = [len(np.where(pred_class == i)[0]) for i in range(nb_classes)]
        # percentage of non-classified lcs
        idx = np.where(np.max(arr_preds, axis=1) < 1 / nb_classes)[0]
        percentage = len(idx) * 100.0 / len(arr_preds)
        list_clf_stats.append(percentage)

        data = np.array(list_clf_stats).reshape(1, -1)
        columns = [f"{prefix}_num_pred_class{i}" for i in range(nb_classes)]
        columns += [f"{prefix}_percentage_non_pred"]

        list_df.append(pd.DataFrame(data, columns=columns))

    df_stats = pd.concat(list_df, axis=1)

    return df_stats



import numpy as np
import pandas as pd
from pathlib import Path
from tabulate import tabulate
from ..utils import data_utils as du
from ..utils import logging_utils as lu
from ..utils import performance_utils as pu


"""
Useful functions for metrics quoted in SuperNNova

"""


def select_df(df, sel_criteria, data=None):
    """Select a subsample of a pandas Dataframe
    Valid only for one to two selection criteria

    Args:
        df (Pandas.Dataframe)
        sel_criteria (list): selection criteria
        data (optional): if source_data must be override
    """
    if data:
        df = df[df["source_data"].str.contains(data)]
    else:
        df = df

    n_crit = len(sel_criteria)

    if n_crit == 1:
        df_sel = df[df["model_name_noseed"].str.contains(sel_criteria[0])]
        print(f"{sel_criteria[0]}")
    elif n_crit == 2:
        df_sel = df[
            (df["model_name_noseed"].str.contains(sel_criteria[0]))
            & df["model_name_noseed"].str.contains(sel_criteria[1])
        ]
        print(f"{sel_criteria[0]} {sel_criteria[1]} ")

    return df_sel


def acc_auc_df(df, model_names_list, data=None):
    df_sel = df[df["model_name_noseed"].isin(model_names_list)]
    if data:
        df_sel = df_sel[df_sel["source_data"].str.contains(data)]
    else:
        df_sel = df_sel
    df.round(
        {
            "all_accuracy_mean": 2,
            "all_accuracy_std": 2,
            "all_auc_mean": 4,
            "all_auc_std": 4,
        }
    )
    nice_df_print(
        df_sel,
        keys=[
            "model_name_noseed",
            "all_accuracy_mean",
            "all_accuracy_std",
            "all_auc_mean",
            "all_auc_std",
        ],
    )


def get_metric_ranges(df, sel_criteria, metric="all_accuracy", round_output=2):
    df_sel = select_df(df, sel_criteria)
    mean_metric = np.round(df_sel[f"{metric}_mean"].mean(), round_output)
    std_metric = np.round(df_sel[f"{metric}_mean"].std())
    top_models = df_sel.nlargest(3, f"{metric}_mean").round(round_output)

    print(f"mean of {mean_metric} \\pm {std_metric}")
    print("top models:")
    nice_df_print(
        top_models, keys=["model_name_noseed", f"{metric}_mean", f"{metric}_std"]
    )

    return df_sel


def nice_df_print(df, keys="keys"):
    if keys != "keys":
        df = df[keys]
    print(tabulate(df, headers=keys, tablefmt="simple", showindex=False))


def get_delta_metrics(df_stats, settings):
    """Difference between models in SuperNNova paper.

    BEWARE: selection hard coded

    Args:
        df (pandas.DataFrame): dataframe containing summary stats
        settings (ExperimentSettings): custom class to hold hyperparameters
    Returns:
        df (pandas.DataFrame): dataframe containing delta metrics
    """

    list_metrics = [
        "all_accuracy_mean",
        "mean_all_class0_std_dev_mean",
        "all_entropy_mean",
    ]

    list_df_delta = []

    # Build pairs of models based on available data
    list_config = df_stats[["model_name_noseed", "source_data"]].values.tolist()
    list_pair = []
    # Look for photometry + saltfit
    list_pair += [
        (
            c,
            [c[0].replace("saltfit_DF_1.0", "photometry_DF_0.43"), "photometry"],
            "salt_phot",
        )
        for c in list_config
        if "saltfit_DF_1.0" in c[0]
        and "photometry" in c[1]
        and [c[0].replace("saltfit_DF_1.0", "photometry_DF_0.43"), "photometry"]
        in list_config
    ]

    # Look for saltfit + saltfit
    list_pair += [
        (c, [c[0].replace("saltfit_DF_0.5", "saltfit_DF_1.0"), "saltfit"], "salt_salt")
        for c in list_config
        if "saltfit_DF_0.5" in c[0]
        and "saltfit" in c[1]
        and [c[0].replace("saltfit_DF_0.5", "saltfit_DF_1.0"), "saltfit"] in list_config
    ]

    # Look for photometry + photometry
    list_pair += [
        (
            c,
            [c[0].replace("photometry_DF_0.43", "photometry_DF_1.0"), "photometry"],
            "phot_phot",
        )
        for c in list_config
        if "photometry_DF_0.43" in c[0]
        and "photometry" in c[1]
        and [c[0].replace("photometry_DF_0.43", "photometry_DF_1.0"), "photometry"]
        in list_config
    ]

    for (c0, c1, delta_type) in list_pair:
        df1 = df_stats[
            (df_stats["model_name_noseed"] == c0[0])
            & (df_stats["source_data"] == c0[1])
        ].reset_index(drop=True)
        df2 = df_stats[
            (df_stats["model_name_noseed"] == c1[0])
            & (df_stats["source_data"] == c1[1])
        ].reset_index(drop=True)

        df_delta = (df1[list_metrics] - df2[list_metrics]).reset_index(drop=True)
        df_delta = df_delta.add_suffix("_delta")
        df_delta["model_name_left"] = df1["model_name_noseed"]
        df_delta["model_name_right"] = df2["model_name_noseed"]
        df_delta["delta_type"] = delta_type

        list_df_delta.append(df_delta)

    if len(list_df_delta) != 0:

        orderded_columns = ["model_name_left", "model_name_right", "delta_type"] + list(
            map(lambda x: f"{x}_delta", list_metrics)
        )

        df_delta = pd.concat(list_df_delta).reset_index(drop=True)
        df_delta = df_delta[orderded_columns]

        df_delta.to_csv(
            Path(settings.stats_dir) / "summary_stats_delta.csv", index=False
        )
    else:
        df_delta = pd.DataFrame(
            columns=[
                "model_name_left",
                "model_name_right",
                "delta_type",
                "all_accuracy_mean_delta",
                "mean_all_class0_std_dev_mean_delta",
                "all_entropy_mean_delta",
            ]
        )

    # Also look for difference between OOD / not OOD
    orderded_columns = ["model_name_noseed"]

    for OOD in du.OOD_TYPES:
        df_stats[f"{OOD}_delta_entropy"] = (
            df_stats["all_entropy_mean"] - df_stats[f"all_{OOD}_entropy_mean"]
        )
        df_stats[f"{OOD}_delta_std_dev"] = (
            df_stats["mean_all_class0_std_dev_mean"]
            - df_stats[f"mean_all_{OOD}_class0_std_dev_mean"]
        )

        orderded_columns += [f"{OOD}_delta_entropy", f"{OOD}_delta_std_dev"]

    df_delta_ood = df_stats[orderded_columns]
    df_delta_ood.to_csv(
        Path(settings.stats_dir) / "summary_stats_delta_OOD.csv", index=False
    )
    return df_delta, df_delta_ood


def create_accuracy_latex_tables(df, settings):
    """Latex accuracy tables for paper.

    BEWARE: Selection is hardcoded

    Args:
        df (pandas.DataFrame) : summary statistics df
        settings (ExperimentSettings): custom class to hold hyperparameters
    """

    tables_to_plot = {
        "accuracies_biclass.tex": {
            "list_criteria": [
                [
                    "vanilla",
                    "_CLF_2",
                    "saltfit",
                    "_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C",
                ],
                [
                    "vanilla",
                    "_CLF_2",
                    "photometry",
                    "_DF_0.43_N_global_lstm_32x2_0.05_128_True_mean_C",
                ],
                [
                    "vanilla",
                    "_CLF_2",
                    "photometry",
                    "_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C",
                ],
            ],
            "list_title": [
                "SALT2 fitted dataset",
                "$43 \\%$ of complete dataset",
                "Complete dataset",
            ],
        },
        "accuracies_multiclass.tex": {
            "list_criteria": [
                [
                    "vanilla",
                    "_CLF_3",
                    "photometry",
                    "_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C",
                ],
                [
                    "vanilla",
                    "_CLF_7",
                    "photometry",
                    "_DF_1.0_N_global_lstm_32x2_0.05_128_True_mean_C",
                ],
            ],
            "list_title": ["Ternary classification", "Seven-way classification"],
        },
    }

    for tabl in tables_to_plot.keys():
        to_write = []
        for i in range(len(tables_to_plot[tabl]["list_criteria"])):
            df_sel = df.copy()
            for j in range(len(tables_to_plot[tabl]["list_criteria"][i])):
                df_sel = df_sel[
                    df_sel["model_name_noseed"].str.contains(
                        tables_to_plot[tabl]["list_criteria"][i][j]
                    )
                ]
            df_sel = df_sel[
                df_sel["source_data"].str.contains(
                    tables_to_plot[tabl]["list_criteria"][i][2]
                )
            ]
            to_write.append(
                pu.create_latex_accuracy_singletable(
                    df_sel,
                    "".join(tables_to_plot[tabl]["list_criteria"][i]),
                    tables_to_plot[tabl]["list_title"][i],
                )
            )

        with open(f"{settings.latex_dir}/{tabl}", "w") as tf:
            for i in range(len(tables_to_plot[tabl]["list_criteria"])):
                tf.write(to_write[i])


def print_contamination(df, sel_criteria, settings, data="saltfit"):
    df_sel = select_df(df, sel_criteria, data=data)
    df_sel = df_sel.round(2)
    print(lu.str_to_bluestr(f"Contamination and efficiency {sel_criteria}"))
    for sntype in [k for k in settings.sntypes.keys() if k != 101]:
        key_list = [
            f"all_contamination_{sntype}_mean",
            f"all_contamination_{sntype}_std",
        ]
        df_sel[f"str_all_contamination_{sntype}"] = df_sel[key_list].apply(
            lambda x: " \\pm ".join(x.map(str)), axis=1
        )
        key_list = [f"0_contamination_{sntype}_mean", f"0_contamination_{sntype}_std"]
        df_sel[f"str_0_contamination_{sntype}"] = df_sel[key_list].apply(
            lambda x: " \\pm ".join(x.map(str)), axis=1
        )

    print("all")
    keys_to_use = ["model_name_noseed"] + [
        k
        for k in df_sel.keys()
        if "str_all_contamination_" in k or "all_efficiency_" in k or "all_purity_" in k
    ]
    nice_df_print(df_sel, keys=keys_to_use)
    print("Peakmjd")
    keys_to_use = ["model_name_noseed"] + [
        k
        for k in df_sel.keys()
        if "str_0_contamination_" in k or "0_efficiency_" in k or "0_purity_" in k
    ]
    nice_df_print(df_sel, keys=keys_to_use)



import os
import h5py
import torch
import numpy as np
import pandas as pd
import torch.nn as nn
from tqdm import tqdm
from pathlib import Path
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from ..utils import data_utils as du
from ..utils import logging_utils as lu
from ..utils import training_utils as tu
from ..utils.visualization_utils import FILTER_COLORS, ALL_COLORS, LINE_STYLE

plt.switch_backend("agg")


def get_predictions(settings, dict_rnn, X, target, OOD=None):

    list_data = [(X.copy(), target)]

    _, X_tensor, *_ = tu.get_data_batch(list_data, [0], settings, OOD=OOD)

    if settings.use_cuda:
        X_tensor.cuda()

    seq_len = X_tensor.shape[0]
    d_pred = {key: {"prob": []} for key in dict_rnn}

    # Loop over light curve time steps to obtain prediction for each time step
    for i in range(1, seq_len + 1):
        # Slice along the time step dimension
        X_slice = X_tensor[:i]

        # Apply rnn to obtain prediction
        for model_type, rnn in dict_rnn.items():

            n = settings.num_inference_samples if "variational" in model_type else 1
            new_size = (X_slice.size(0), n, X_slice.size(2))

            if "bayesian" in model_type:

                # Loop over num samples to obtain predictions
                list_out = [
                    rnn(X_slice.expand(new_size))
                    for i in range(settings.num_inference_samples)
                ]
                out = torch.cat(list_out, dim=0)
                # Apply softmax to obtain a proba
                pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()
            elif "swag" in model_type:

                list_out = []
                # Loop over num samples to obtain predictions
                for i in range(settings.num_inference_samples):
                    sample_model = rnn.sample()
                    list_out.append(sample_model(X_slice.expand(new_size)))
                out = torch.cat(list_out, dim=0)
                # Apply softmax to obtain a proba
                pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

            else:
                out = rnn(X_slice.expand(new_size))
                # Apply softmax to obtain a proba
                pred_proba = nn.functional.softmax(out, dim=-1).data.cpu().numpy()

            # Add to buffer list
            d_pred[model_type]["prob"].append(pred_proba)

    # Stack
    for key in dict_rnn.keys():
        arr_proba = np.stack(d_pred[key]["prob"], axis=0)
        d_pred[key]["prob"] = arr_proba  # arr_prob is (T, num_samples, 2)
        d_pred[key]["median"] = np.median(arr_proba, axis=1)
        d_pred[key]["perc_16"] = np.percentile(arr_proba, 16, axis=1)
        d_pred[key]["perc_84"] = np.percentile(arr_proba, 84, axis=1)
        d_pred[key]["perc_2"] = np.percentile(arr_proba, 2, axis=1)
        d_pred[key]["perc_98"] = np.percentile(arr_proba, 98, axis=1)

    return d_pred, X_tensor.squeeze().detach().cpu().numpy()


def plot_predictions(
    settings,
    d_plot,
    SNID,
    redshift,
    peak_MJD,
    target,
    arr_time,
    d_pred,
    OOD,
    SNtype_str,
):

    plt.figure(figsize=(15, 10))
    gs = gridspec.GridSpec(1 + len(d_pred.keys()), 1)
    # Plot the lightcurve
    ax = plt.subplot(gs[0])
    for n, flt in enumerate(d_plot.keys()):
        flt_time = d_plot[flt]["MJD"]
        # Only plot a time series if it's non empty
        if len(flt_time) > 0:
            flux = d_plot[flt]["FLUXCAL"]
            fluxerr = d_plot[flt]["FLUXCALERR"]
            ax.errorbar(
                flt_time,
                flux,
                yerr=fluxerr,
                fmt="o",
                label=f"Filter {flt}",
                color=FILTER_COLORS[flt]
                if flt in FILTER_COLORS.keys()
                else ALL_COLORS[n],
            )
    ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    ax.set_ylabel("FLUXCAL")
    ylim = ax.get_ylim()

    if settings.data_testing:
        ax.set_title(f"ID: {SNID}")
    elif OOD is not None:
        ax.set_title(f"OOD {OOD} ID: {SNID}")
    else:
        ax.set_title(SNtype_str + f" (ID: {SNID}, redshift: {redshift:.3g})")
        # Add PEAKMJD
        if (
            OOD is None
            and not settings.data_testing
            and arr_time.min() < peak_MJD
            and peak_MJD < arr_time.max()
        ):
            ax.plot([peak_MJD, peak_MJD], ylim, "k--", label="Peak MJD")

    d_keys = {
        "vanilla photometry": "LSTM",
        "variational photometry": "MC Dropout",
        "bayesian photometry": "BNN",
        "swag photometry": "SWAG",
    }

    # Plot the classifications
    for idx, key in enumerate(d_pred.keys()):

        ax = plt.subplot(gs[idx + 1])
        ax.set_ylim(0, 1)

        for class_prob in range(settings.nb_classes):
            color = ALL_COLORS[class_prob]
            linestyle = LINE_STYLE[class_prob]
            label = du.sntype_decoded(class_prob, settings)
            if class_prob != 0 and settings.nb_classes < 3:
                label = "non-Ia"

            if len(d_pred) > 1:
                label += f" {d_keys.get(key, key)}"

            ax.plot(
                arr_time,
                d_pred[key]["median"][:, class_prob],
                color=color,
                linestyle=linestyle,
                label=label,
            )
            ax.fill_between(
                arr_time,
                d_pred[key]["perc_16"][:, class_prob],
                d_pred[key]["perc_84"][:, class_prob],
                color=color,
                alpha=0.4,
            )
            ax.fill_between(
                arr_time,
                d_pred[key]["perc_2"][:, class_prob],
                d_pred[key]["perc_98"][:, class_prob],
                color=color,
                alpha=0.2,
            )

        ax.set_xlabel("Time (MJD)")
        ax.set_ylabel("classification probability")
        # Add PEAKMJD
        if (
            OOD is None
            and not settings.data_testing
            and arr_time.min() < peak_MJD
            and peak_MJD < arr_time.max()
        ):
            ax.plot([peak_MJD, peak_MJD], [0, 1], "k--", label="Peak MJD")
        ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)

    prefix = f"OOD_{OOD}_" if OOD is not None else ""

    if len(settings.model_files) > 1:
        fig_path = f"{settings.figures_dir}/{prefix}multi_model_early_prediction"
        fig_name = f"{prefix}multi_model_{SNID}.png"
    elif len(settings.model_files) == 1:
        parent_dir = Path(settings.model_files[0]).parent.name
        fig_path = f"{settings.lightcurves_dir}/{parent_dir}/{prefix}early_prediction"
        fig_name = f"{parent_dir}_{prefix}class_pred_with_lc_{SNID}.png"
    else:
        fig_path = f"{settings.lightcurves_dir}/{settings.pytorch_model_name}/{prefix}early_prediction"
        fig_name = (
            f"{settings.pytorch_model_name}_{prefix}class_pred_with_lc_{SNID}.png"
        )
    Path(fig_path).mkdir(parents=True, exist_ok=True)
    plt.tight_layout()
    plt.savefig(Path(fig_path) / fig_name)
    plt.clf()
    plt.close()


def make_early_prediction(settings, nb_lcs=1, do_gifs=False):
    """Load model corresponding to settings
    or (if specified) load a list of models.

    - Show evolution of classification for one time-step, then 2, up to all of the lightcurve
    - For Bayesian models, show uncertainty in the prediction
    - Figures are save in the figures repository

    Args:
        settings: (ExperimentSettings) custom class to hold hyperparameters
        int (nb_lcs): number of light-curves to plot, default is 1
    """
    settings.random_length = False
    settings.random_redshift = False

    # Load the test data
    list_data_test = tu.load_HDF5(settings, test=True)

    # Load features list
    file_name = f"{settings.processed_dir}/database.h5"
    with h5py.File(file_name, "r") as hf:
        features = hf["features"][settings.idx_features].astype(str)

    # Load RNN model
    dict_rnn = {}
    if settings.model_files is None:
        settings.model_files = [f"{settings.rnn_dir}/{settings.pytorch_model_name}.pt"]
    else:
        # check if the model files are there
        tmp_not_found = [m for m in settings.model_files if not os.path.exists(m)]
        if len(tmp_not_found) > 0:
            print(lu.str_to_redstr(f"Files not found {tmp_not_found}"))
            tmp_model_files = [m for m in settings.model_files if os.path.exists(m)]
            settings.model_files = tmp_model_files

    for model_file in settings.model_files:
        if "variational" in model_file:
            settings.model = "variational"
        if "vanilla" in model_file:
            settings.model = "vanilla"
        if "bayesian" in model_file:
            settings.model = "bayesian"
        if "swag" in model_file:
            settings.model = "swag"
        rnn = tu.get_model(settings, len(settings.training_features))
        rnn_state = torch.load(
            model_file,
            map_location=lambda storage, loc: storage,
            weights_only=False,
        )
        if isinstance(rnn_state, torch.nn.Module):
            # TODO(anaismoller) fix this
            rnn_state = rnn_state.state_dict()
            module_state = {k: v for (k, v) in rnn_state.items() if "module." in k}
            non_module_state = {
                k: v for k, v in rnn_state.items() if "module." not in k
            }
            rnn.load_state_dict(module_state, strict=False)
            for key in list(non_module_state.keys()):
                val = non_module_state[key]
                obj = getattr(rnn, key)
                obj.data = val
        else:
            rnn.load_state_dict(rnn_state, strict=False)
        rnn.to(settings.device)
        rnn.eval()
        name = (
            f"{settings.model} photometry"
            if "photometry" in model_file
            else f"{settings.model} salt"
        )
        dict_rnn[name] = rnn

    # load SN info
    SNinfo_df = du.load_HDF5_SNinfo(settings)

    # Loop over data to plot prediction
    if settings.plot_file:
        # plot only lcs with  SNID in csv file
        if os.path.exists(settings.plot_file):
            tmp = pd.read_csv(settings.plot_file)
            list_to_plot = tmp.SNID.astype(str).tolist()
            subset_to_plot = filter(lambda x: x[2] in list_to_plot, list_data_test)
            subset_to_plot = list(subset_to_plot)
        else:
            lu.print_red(f"Not a valid file --plot_file {settings.plot_file}")
            lu.print_yellow("Plotting 2 random lcs")
            # randomly select lcs to plot
            list_entries = np.random.randint(0, high=len(list_data_test), size=2)
            subset_to_plot = [list_data_test[i] for i in list_entries]
    else:
        # randomly select lcs to plot
        list_entries = np.random.randint(0, high=len(list_data_test), size=nb_lcs)
        subset_to_plot = [list_data_test[i] for i in list_entries]

    for X, target, SNID, _, X_ori in tqdm(subset_to_plot, ncols=100):

        try:
            redshift = SNinfo_df[SNinfo_df["SNID"] == SNID]["SIM_REDSHIFT_CMB"].values[
                0
            ]
            peak_MJD = SNinfo_df[SNinfo_df["SNID"] == SNID]["PEAKMJDNORM"].values[0]
            SNtype_str = settings.sntypes[
                str(SNinfo_df[SNinfo_df["SNID"] == SNID][settings.sntype_var].values[0])
            ]
        except (IndexError, KeyError):
            redshift = 0.0
            peak_MJD = 0.0
            SNtype_str = "Not found"

        # Prepare plotting data in a dict
        d_plot = {
            flt: {"FLUXCAL": [], "FLUXCALERR": [], "MJD": []}
            for flt in settings.list_filters
        }
        for OOD in [None]:  # + du.OOD_TYPES: # uncomment to plot OOD
            with torch.no_grad():
                d_pred, X_normed = get_predictions(
                    settings, dict_rnn, X, target, OOD=OOD
                )
            # X here has been normalized. We unnormalize X
            try:
                X_unnormed = tu.unnormalize_arr(X_normed, settings)

                # TODO: IMPROVE
                df_temp = pd.DataFrame(data=X_unnormed, columns=features)
                arr_time = np.cumsum(df_temp.delta_time.values)
                df_temp["time"] = arr_time
                for flt in settings.list_filters:
                    non_zero = np.where(
                        ~np.isclose(df_temp[f"FLUXCAL_{flt}"].values, 0, atol=1e-2)
                    )[0]
                    d_plot[flt]["FLUXCAL"] = df_temp[f"FLUXCAL_{flt}"].values[non_zero]
                    d_plot[flt]["FLUXCALERR"] = df_temp[f"FLUXCALERR_{flt}"].values[
                        non_zero
                    ]
                    d_plot[flt]["MJD"] = arr_time[non_zero]
                plot_predictions(
                    settings,
                    d_plot,
                    SNID,
                    redshift,
                    peak_MJD,
                    target,
                    arr_time,
                    d_pred,
                    OOD,
                    SNtype_str,
                )

                # use to create GIFs
                if not OOD:
                    if do_gifs:
                        plot_gif(
                            settings,
                            df_temp,
                            SNID,
                            redshift,
                            peak_MJD,
                            target,
                            arr_time,
                            d_pred,
                        )
            except (IndexError, ValueError) as e:
                lu.print_red(f"SNID {SNID} only has {len(X)} measurement, not plotting ({e})")

    lu.print_green("Finished plotting lightcurves and predictions ")


def plot_gif(settings, df_plot, SNID, redshift, peak_MJD, target, arr_time, d_pred):
    """Create GIFs for classification"""

    def plot_image_for_gif(fig, gs, df_plot, d_pred, time, SNtype):

        # Plot the lightcurve
        ax = plt.subplot(gs[0])
        # Used to keep the limits constant
        flux_max = max(df_plot[[k for k in df_plot.keys() if "FLUXCAL_" in k]].max())
        flux_min = min(df_plot[[k for k in df_plot.keys() if "FLUXCAL_" in k]].min())
        ax.set_ylim(flux_min - 5, flux_max + 5)
        ax.set_xlim(-0.5, max(df_plot["time"]) + 2)

        # slice for gif
        df_sel = df_plot[df_plot["time"] <= time]
        for n, flt in enumerate(settings.list_filters):
            ax.errorbar(
                df_sel["time"],
                df_sel[f"FLUXCAL_{flt}"],
                yerr=df_sel[f"FLUXCALERR_{flt}"],
                fmt="o",
                label=f"Filter {flt}",
                color=FILTER_COLORS[flt]
                if flt in FILTER_COLORS.keys()
                else ALL_COLORS[n],
            )

        ax.set(
            xlabel="",
            ylabel="flux",
            title=f"{SNtype} (ID: {SNID}, redshift: {redshift:.3g})",
        )

        # Plot the classifications
        ax = plt.subplot(gs[1])
        ax.clear()
        ax.set_ylim(0, 1)
        ax.set_xlim(-0.5, max(df_plot["time"]) + 2)
        # select classification of same length
        for idx, key in enumerate(d_pred.keys()):

            for class_prob in range(settings.nb_classes):
                color = ALL_COLORS[class_prob + idx * settings.nb_classes]
                linestyle = LINE_STYLE[class_prob]
                label = du.sntype_decoded(class_prob, settings)
                if len(d_pred) > 1:
                    label += f" {key}"

                ax.plot(
                    arr_time[: len(df_sel)],
                    d_pred[key]["median"][:, class_prob][: len(df_sel)],
                    color=color,
                    linestyle=linestyle,
                    label=label,
                )
                ax.fill_between(
                    arr_time[: len(df_sel)],
                    d_pred[key]["perc_16"][:, class_prob][: len(df_sel)],
                    d_pred[key]["perc_84"][:, class_prob][: len(df_sel)],
                    color=color,
                    alpha=0.4,
                )
                ax.fill_between(
                    arr_time[: len(df_sel)],
                    d_pred[key]["perc_2"][:, class_prob][: len(df_sel)],
                    d_pred[key]["perc_98"][:, class_prob][: len(df_sel)],
                    color=color,
                    alpha=0.2,
                )
        ax.set_ylabel("classification probability")
        ax.set_xlabel("time")

        # Used to return the plot as an image rray
        fig.canvas.draw()  # draw the canvas, cache the renderer
        image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
        image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))

        from PIL import Image

        im = Image.fromarray(image)
        # make background transparent
        # im = im.convert('RGB')
        # im =pops.invert(im)
        image = im

        return image

    fig = plt.figure()
    gs = gridspec.GridSpec(2, 1)
    SNtype = du.sntype_decoded(target, settings)

    fig_path = f"{settings.lightcurves_dir}/{settings.pytorch_model_name}/gif"
    fig_name = f"{settings.pytorch_model_name}_class_pred_with_lc_{SNID}.gif"
    Path(fig_path).mkdir(parents=True, exist_ok=True)
    arr_images = [
        plot_image_for_gif(fig, gs, df_plot, d_pred, time, SNtype) for time in arr_time
    ]
    arr_images[0].save(
        str(Path(fig_path) / fig_name),
        save_all=True,
        append_images=arr_images,
        loop=5,
        duration=200,
    )



from time import time
from colorama import init, Fore, Style
from contextlib import ContextDecorator


class timer(ContextDecorator):
    """
    Args:
        msg (str): string message to display when timing
        verbose (bool): If True, will print the timing.
            Default: ``False``
    """

    def __init__(self, msg):

        self.msg = msg

    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, type, value, traceback):
        total = time() - self.start
        print()
        print_green(f"{self.msg}", f"{pretty_time(total)}")


def pretty_time(orig_seconds):
    """Transforms seconds into a string with days, hours, minutes and seconds."""
    days, seconds = divmod(round(orig_seconds), 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)

    out = []
    if days > 0:
        out.append(f"{days}d")
    if hours > 0:
        out.append(f"{hours}h")
    if minutes > 0:
        out.append(f"{minutes}m")
    if seconds > 0:
        out.append(f"{seconds}s")
    else:
        if out:
            s = ""
        elif orig_seconds == 0:
            s = "0s"
        else:
            s = "<0s"
        out.append(s)

    return "".join(out)


def print_bright(s):

    init()
    print(Style.BRIGHT + s + Style.RESET_ALL)


def print_green(info, value="", verbose=True):
    if verbose is False:
        return

    print(Fore.GREEN + "[%s] " % info + Style.RESET_ALL + str(value))


def print_yellow(info, value="", verbose=True):
    if verbose is False:
        return

    print(Fore.YELLOW + "[%s] " % info + Style.RESET_ALL + str(value))


def print_red(info, value="", verbose=True):
    if verbose is False:
        return
    print(Fore.RED + "[%s] " % info + Style.RESET_ALL + str(value))


def print_blue(info, value="", verbose=True):
    if verbose is False:
        return
    print(Fore.BLUE + "[%s] " % info + Style.RESET_ALL + str(value))


def str_to_brightstr(string):

    return Style.BRIGHT + "%s" % string + Style.RESET_ALL


def str_to_redstr(string):

    return Fore.RED + "%s" % string + Style.RESET_ALL


def str_to_bluestr(string):

    return Fore.BLUE + "%s" % string + Style.RESET_ALL


def str_to_yellowstr(string):

    return Fore.YELLOW + "%s" % string + Style.RESET_ALL


def str_to_greenstr(string):

    return Fore.GREEN + "%s" % string + Style.RESET_ALL
